
from __future__ import annotations

import array
import logging
from typing import Optional, TYPE_CHECKING, Any, Union

from jpype.types import *

from dragodis.exceptions import NotExistError
from dragodis.ghidra.instruction import GhidraInstruction
from dragodis.interface import Line, LineType, CommentType

# Used for typing.
if TYPE_CHECKING:
    from dragodis.ghidra.flat import GhidraFlatAPI
    import ghidra
    # A "CodeUnit" that line takes in is a subclass of CodeUnit.
    CodeUnit = Union[
        ghidra.program.model.listing.CodeUnit,
        ghidra.program.model.listing.Instruction,
        ghidra.program.model.listing.Data,
    ]

logger = logging.getLogger(__name__)


class GhidraLine(Line):

    _comment_type_map = {
        CommentType.eol: 0,          # EOL_COMMENT
        CommentType.anterior: 1,     # PRE_COMMENT
        CommentType.posterior: 2,    # POST_COMMENT
        CommentType.plate: 3,        # PLATE_COMMENT
        CommentType.repeatable: 4,   # REPEATABLE_COMMENT
    }

    # Maps Ghidra DataType to LineType
    _data_type_map = {
        "byte": LineType.byte,
        "word": LineType.word,
        "short": LineType.word,
        "dword": LineType.dword,
        "int": LineType.dword,
        "qword": LineType.qword,
        "int16": LineType.oword,
        "float": LineType.float,
        "float4": LineType.float,
        "double": LineType.double,
        # TODO: Support pointer types?
        #   (Also, this could be qword if 64bit)
        "pointer": LineType.dword,
        "string": LineType.string,
        "string-utf8": LineType.string,
        "TerminatedCString": LineType.string,
        "unicode": LineType.string16,
        "TerminatedUnicode": LineType.string16,
        "unicode32": LineType.string32,
        "TerminatedUnicode32": LineType.string32,
        "Alignment": LineType.align,
        # TODO: struct
    }
    # Inverse _data_type_map with most recent item as value for many-to-one scenerio.
    # (We can do this because Python 3 dictionaries are ordered!)
    _data_type_map_inv = {}
    for k, v in _data_type_map.items():
        if v not in _data_type_map_inv:
            _data_type_map_inv[v] = k

    def __init__(self, api: GhidraFlatAPI, code_unit: "CodeUnit"):
        super().__init__(api)
        self._ghidra = api
        self._code_unit = code_unit

    @property
    def _addr_obj(self) -> "ghidra.program.model.address.Address":
        return self._code_unit.getAddress()

    @property
    def _next_defined(self) -> "GhidraLine":
        """
        Gets the next defined line. (helps to simulate IDA's next_head())
        """
        line = self.next
        while line.type == LineType.undefined:
            line = line.next
        return line

    @property
    def address(self) -> int:
        return self._addr_obj.getOffset()

    @property
    def data(self) -> bytes:
        # NOTE: Need to cast returned array.array from Ghidra, so we can
        # get a Python 3 version of array.array that has the tobytes() function.
        # TODO: Undo this when we use Jpype?
        if not self.is_loaded:
            return b""
        else:
            return array.array("b", self._code_unit.getBytes()).tobytes()

    @data.setter
    def data(self, new_data: bytes):
        if not self.is_loaded:
            return
        type_ = self.type
        self.undefine()
        self._ghidra._flatapi.setBytes(self._addr_obj, new_data)
        self.type = type_

    def get_comment(self, comment_type=CommentType.eol) -> Optional[str]:
        return self._code_unit.getComment(self._comment_type_map[comment_type])

    @property
    def name(self) -> Optional[str]:
        # NOTE: getLabel() is a shortcut for getPrimarySymbol().getName()
        return self._code_unit.getLabel()

    @name.setter
    def name(self, new_name: str):
        # Sanitize name.
        if new_name:
            new_name = new_name.replace(" ", "")

        # First get the label at the address (if it exists), then we need to either
        # modify/delete the existing label, or create a new label.
        # If we attempt to just call createLabel() Ghidra will stack the label along with
        # the other.
        label = self._code_unit.getPrimarySymbol()  # Ghidra calls labels, "symbols" for legacy reasons.
        if label:
            # First see if name is the same.
            if label.getName() == new_name:
                return
            # Modify existing label
            if new_name:
                # TODO: Ghidra throws a DuplicateNameException if the name exists somewhere
                #   else instead of adding a prefix.
                label.setName(new_name, label.getSource())
            # Remove existing label.
            else:
                label.delete()

        # Set new label.
        elif new_name:
            label = self._ghidra._flatapi.createLabel(self._addr_obj, new_name, True)
            if not label:
                raise ValueError(f"Failed to set {hex(self.address)} with name: {new_name}")

    @property
    def next(self) -> Optional["GhidraLine"]:
        next_code_unit = self._ghidra._listing.getCodeUnitAfter(self._addr_obj)
        if next_code_unit:
            return GhidraLine(self._ghidra, next_code_unit)

    @property
    def prev(self) -> Optional["GhidraLine"]:
        prev_code_unit = self._ghidra._listing.getCodeUnitBefore(self._addr_obj)
        if prev_code_unit:
            return GhidraLine(self._ghidra, prev_code_unit)

    def set_comment(self, comment: Optional[str], comment_type=CommentType.eol):
        # Ghidra clears comment when null
        if not comment:
            comment = None

        self._code_unit.setComment(self._comment_type_map[comment_type], comment)

    @property
    def size(self) -> int:
        return self._code_unit.getLength()

    # TODO: Need to look into reevaluating what a line type can contain to better support ghidra
    @property
    def type(self) -> LineType:
        code_unit_type = str(self._code_unit.getClass())

        if "Instruction" in code_unit_type:
            return LineType.code

        data_type = self._code_unit.getDataType().getName()

        if "struct" in data_type:  # FIXME
            return LineType.struct

        # If data type is undefined, check if data is unloaded by checking if it has a value.
        if data_type == "undefined" and self._code_unit.getValue() is None:
            return LineType.unloaded

        if data_type.startswith("undefined"):
            return LineType.undefined

        try:
            return self._data_type_map[data_type]
        except KeyError:
            raise RuntimeError(f"Unexpected line type at {hex(self.address)}")

    @type.setter
    def type(self, new_type: LineType):
        unsupported = (LineType.tbyte, LineType.pack_real)
        if new_type not in LineType or new_type in unsupported:
            raise ValueError(f"Invalid line type: {new_type}")

        orig_type = self.type
        if orig_type == new_type:
            return  # No change necessary

        if new_type == LineType.undefined:
            self.undefine()
            return

        if new_type == LineType.unloaded:
            # TODO: Should we ever support this?
            raise TypeError(f"Setting a line type to unloaded is not currently supported.")

        if new_type == LineType.code:
            raise NotImplementedError("Setting a line to code is not currently supported.")

        if new_type == LineType.struct:
            raise NotImplementedError("Setting a line type to struct is not currently supported.")

        manager = self._ghidra._program.getDataTypeManager()
        data_type = manager.getDataType(f"/{self._data_type_map_inv[new_type]}")
        if not data_type:
            # FIXME: For some unknown reason somes types don't show up in the DataTypeManager.
            #   Therefore, we are going to initialize it manually.
            #   Determine if we need to do some type of data type initialization or something.
            if new_type == LineType.align:
                from ghidra.program.model.data import AlignmentDataType
                data_type = AlignmentDataType(manager)
            elif new_type == LineType.float:
                from ghidra.program.model.data import FloatDataType
                data_type = FloatDataType(manager)

            if not data_type:
                raise TypeError(f"Invalid line type: {repr(new_type)}")

        # TODO: catch possible exception thrown by Ghidra.
        self.undefine()
        self._code_unit = self._ghidra._flatapi.createData(self._addr_obj, data_type)

    def undefine(self):
        # TODO: catch the "CancelledException" that could be thrown.
        self._ghidra._flatapi.clearListing(self._addr_obj)
        self._code_unit = self._ghidra._listing.getCodeUnitAt(self._addr_obj)

    @property
    def value(self) -> Any:
        line_type = self.type

        if line_type == LineType.code:
            # TODO: Need to figure out generic way to pull x86Instruction and ARMInstruction types from Instruction???
            return GhidraInstruction(self._ghidra, self._code_unit)
        elif line_type == LineType.align:
            return self.data

        value = self._code_unit.getValue()

        # Scalar object
        if hasattr(value, "getUnsignedValue"):
            # NOTE: Right now we are forcing a signed value by calling getUnsignedValue(),
            #   but we may want to selectively use getValue() or getSignedValue().
            value = value.getUnsignedValue()

        # Pointer object
        if hasattr(value, "getOffset"):
            value = value.getOffset()

        return value

    @value.setter
    def value(self, new_value: Any):
        from ghidra.program.model.mem import MemoryAccessException
        try:
            self._set_value(new_value)
        except MemoryAccessException:
            block = self._ghidra._flatapi.getMemoryBlock(self._addr_obj)
            if block.isInitialized():
                # caused by some other problem
                raise

            # initialize the block and try again
            memory = self._ghidra._program.getMemory()
            memory.convertToInitialized(block, 0)
            self._set_value(new_value)

    def _set_value(self, new_value: Any):
        matched_types = LineType.match_type(new_value)
        if not matched_types:
            raise TypeError(f"Unsupported value type: {type(new_value)}")

        type_ = self.type

        # First see if setting this new value would require changing the line type.
        if type_ not in matched_types:
            # Use the first entry of matched types to be the type we change it to.
            new_type = matched_types[0]
            logger.debug(f"Changing line type from {type_} to {new_type}")
            self.type = new_type
            type_ = new_type

        if type_ == LineType.code:
            raise NotImplementedError(f"Setting an instruction is not currently supported.")

        elif type_ == LineType.byte:
            if isinstance(new_value, bytes):
                new_value = new_value[0]
            self._ghidra._flatapi.setByte(self._addr_obj, JByte(new_value))

        elif type_ == LineType.word:
            self._ghidra._flatapi.setShort(self._addr_obj, JShort(new_value))

        elif type_ == LineType.dword:
            self._ghidra._flatapi.setInt(self._addr_obj, JInt(new_value))

        elif type_ == LineType.qword:
            self._ghidra._flatapi.setLong(self._addr_obj, JLong(new_value))

        elif type_ == LineType.oword:
            raise NotImplementedError(f"Setting oword value it not currently supported.")

        elif type_ == LineType.tbyte:
            raise NotImplementedError(f"Setting tbyte value is not currently supported.")

        elif type_ == LineType.float:
            self._ghidra._flatapi.setFloat(self._addr_obj, JFloat(new_value))

        elif type_ == LineType.double:
            self._ghidra._flatapi.setDouble(self._addr_obj, JDouble(new_value))

        elif type_ == LineType.pack_real:
            raise NotImplementedError(f"Setting pack_real value is not currently supported.")

        elif type_ in (LineType.string, LineType.string16, LineType.string32):
            new_value += "\x00"
            encoding = {
                LineType.string: "utf-8",
                LineType.string16: "utf-16-le",
                LineType.string32: "utf-32-le",
            }[type_]
            new_value = new_value.encode(encoding)
            # Ensure there is enough space to fit the new value.
            if len(new_value) > (self._next_defined.address - self.address):
                raise ValueError(f"Encoded string too large to fit on line.")
            # Fill with zero bytes.
            new_value += b"\x00" * (len(self) - len(new_value))
            # For strings, we are going to undefine the value first so ghidra appropriately
            # resets the code unit.
            self.undefine()
            self._ghidra._flatapi.setBytes(self._addr_obj, list(new_value))
            self.type = type_
            return

        elif type_ == LineType.struct:
            raise NotImplementedError(f"Setting struct value is not currently supported.")

        elif type_ == LineType.align:
            self._ghidra._flatapi.setBytes(self._addr_obj, new_value)

        elif type_ == LineType.tail:
            raise NotImplementedError(f"Getting tail value is not currently supported.")

        elif type_ == LineType.undefined:
            if isinstance(new_value, int):
                new_value = bytes([new_value])
            self._ghidra._flatapi.setBytes(self._addr_obj, new_value)

        elif type_ == LineType.unloaded:
            ...  # Nothing needs to be done for unloaded data.



"""
Base interface for dragodis
"""
from __future__ import annotations

# Import available disassemblers so they get registered.
import os
import pathlib
from typing import Union, Type, TYPE_CHECKING

from dragodis import utils
from dragodis.constants import BACKEND_DEFAULT
from dragodis.ghidra import GhidraLocal, GhidraRemote
from dragodis.ida import IDALocal, IDARemote
from dragodis.interface import FlatAPI, BackendDisassembler
from dragodis.interface.types import ProcessorType

IDA = IDALocal if utils.in_ida() else IDARemote

def Ghidra(*args, **kwargs):
    # Need to dynamically provide Ghidra disassembler since detection can be wrong if
    # this module gets imported prematurely by a plugin before pyhidra sets up the interpreter.
    if utils.in_ghidra():
        return GhidraLocal(*args, **kwargs)
    else:
        return GhidraRemote(*args, **kwargs)

if TYPE_CHECKING:
    Ghidra = GhidraLocal if utils.in_ghidra() else GhidraRemote


# Expose flat api when user wants to do typing.
class Disassembler(FlatAPI, BackendDisassembler):
    ...


def _get_class(name: str = None) -> Type[Disassembler]:
    # If name not provided, see if we can detect we are inside a disassembler. If so, use that.
    if not name:
        if utils.in_ida():
            return IDA
        if utils.in_ghidra():
            return Ghidra

    name = name or BACKEND_DEFAULT
    if not name:
        raise ValueError(
            "No disassembler provided. "
            "Please provide disassembler name as an argument or by setting "
            "the environment variable 'DRAGODIS_DISASSEMBLER'."
        )
    if name.lower() == "ida":
        return IDA
    elif name.lower() == "ghidra":
        return Ghidra
    else:
        raise ValueError(f"Not a valid disassembler: {name}")


def open_program(
        file_path: Union[str, pathlib.Path],
        disassembler: str = None,
        processor: Union[ProcessorType, str] = None,
        **config
) -> Disassembler:
    """
    Opens given file in the given disassembler.

    :param file_path: Path to binary file to disassemble.
    :param disassembler: Name of disassembler to use.
        (Defaults to disassembler provided in DRAGODIS_DISASSEMBLER environment variable.)
    :param processor: Processor spec to use.
        (Defaults to auto-detection by underlying disassembler)
    :param config: Arguments to pass to disassembler during instantiation.
        Non-applicable arguments will be ignored.

    :raises NotInstalledError: If the specified disassembler has not been installed.
    :raises ValueError: If the specified disassembler is not supported.
    """
    file_path = pathlib.Path(file_path)
    return _get_class(disassembler)(file_path, processor=processor, **config)




import abc
from typing import Iterable, Optional, Union

import capstone

# TODO: Rename to base?
from dragodis.exceptions import NotExistError
from dragodis.interface.string import String
from dragodis.interface.variable import GlobalVariable
from dragodis.interface.data_type import DataType
from dragodis.interface.function_signature import FunctionSignature
from dragodis.interface.operand import Operand
from dragodis.interface.operand_value import Register, OperandValue
from dragodis.interface.function import Function
from dragodis.interface.flowchart import Flowchart, BasicBlock
from dragodis.interface.instruction import Instruction
from dragodis.interface.memory import Memory
from dragodis.interface.segment import Segment
from dragodis.interface.line import Line
from dragodis.interface.reference import Reference
from dragodis.interface.symbol import Import, Export
from dragodis.interface.types import OperandType, CommentType, ReferenceType


# TODO: Look into using zope.interface instead of abc
class FlatAPI(metaclass=abc.ABCMeta):
    """
    The Flat API is provided to allow easier access to some data that would require more
    steps than necessary to access through the OOP API. The Flat API also provides many of
    the functions that the OOP API is built upon.

    NOTE: This class is designed to be used as a mixin along with a Disassembler class.
    """

    # TODO:
    #   - FlatAPI/Base should handle the caching of different types based on its constructors?
    #       - Or should that be done by the classes themselves? (ie. a from_cache() function)
    #   - FlatAPI/Base should define the classes used to construct Function, Operand, etc.

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)  # forward to Disassembler class
        self.__capstone_dis = None

    def __repr__(self):
        return (
            f"<Disassembler "
            f"{self.name}:{self.processor_name}:{self.bit_size}-{'BE' if self.is_big_endian else 'LE'}>"
        )

    @property
    def _capstone_dis(self) -> capstone.Cs:
        """
        Obtains capstone disassembler.
        This can be used to help fill in for missing feature sets of the backend disassembler.
        """
        if not self.__capstone_dis:
            if self.processor_name == "x86":
                arch = capstone.CS_ARCH_X86
                mode = {
                    16: capstone.CS_MODE_16,
                    32: capstone.CS_MODE_32,
                    64: capstone.CS_MODE_64,
                }[self.bit_size]
            elif self.processor_name == "ARM":
                arch = {
                    32: capstone.CS_ARCH_ARM,
                    64: capstone.CS_ARCH_ARM64,
                }[self.bit_size]
                # TODO: Support thumb mode.
                mode = capstone.CS_MODE_ARM
            else:
                raise NotImplementedError(f"Unsupported processor: {self.processor_name}")
            self.__capstone_dis = capstone.Cs(arch, mode)
            self.__capstone_dis.detail = True
        return self.__capstone_dis

    @property
    @abc.abstractmethod
    def processor_name(self) -> str:
        """
        Returns the name of the processor.
        """

    @property
    @abc.abstractmethod
    def compiler_name(self) -> str:
        """
        Returns the name of the compiler.
        """

    @property
    @abc.abstractmethod
    def bit_size(self) -> int:
        """
        Returns the address bit size of the detected processor.
        """

    @property
    @abc.abstractmethod
    def is_big_endian(self) -> bool:
        """
        Whether the processor is big endian.
        (Assumed to be little endian if False)
        """

    @property
    @abc.abstractmethod
    def entry_point(self) -> Optional[int]:
        """
        The address for the OEP (Original Entry Point) or general starting address for the program (if applicable).
        NOTE: This is usually the "AddressOfEntryPoint" listing. Use .exports to get other possible entry points.

        :yields: Address of the entry point or None if not applicable.
        """

    def is_loaded(self, addr: int) -> bool:
        """
        Determines if the given address exists and is loaded in the sample.

        :param addr: Address to test.
        :return: Whether the address exists and is loaded.
        """
        try:
            return self.get_line(addr).is_loaded
        except NotExistError:
            return False

    @abc.abstractmethod
    def functions(self, start=None, end=None) -> Iterable[Function]:
        """
        Iterates `Function` objects found within the range of the given
        `start` and `end` addresses.

        :param int start: Start address to start iterating functions (defaults to beginning of program)
            If the start address is at the beginning of a function, that function is
            not included. Only functions for which we find the entry point.
        :param int end: End address to end iterating functions (defaults to end of program)
            If end address is within a function, that function may still be counted if the
            entry point is before the end address.
        """

    @abc.abstractmethod
    def get_virtual_address(self, file_offset: int) -> int:
        """
        Obtains virtual address from given file offset.

        :param int file_offset: Offset within underlying source file.
        :raises NotExistsError: If a virtual address doesn't exists for given file offset.
        """

    @abc.abstractmethod
    def get_file_offset(self, addr: int) -> int:
        """
        Obtains offset within underlying source file from given virtual address.

        :param int addr: Linear address.
        :raises NotExistsError: If a file offset doesn't exists for given address.
        """

    # TODO: Move these to some type of Memory object?

    @abc.abstractmethod
    def get_byte(self, addr: int) -> int:
        """
        Returns the one byte value at ``addr``.

        Must throw a NotExistError exception if the byte cannot be retrieved.

        :param int addr: Address where the byte is located
        :return: One byte value
        :rtype: int
        """

    @abc.abstractmethod
    def get_bytes(self, addr: int, length: int, default: int = None) -> bytes:
        """
        Returns the raw bytes at the indicated address.

        If the address is invalid, this must throw a NotExistError exception.

        :param addr: Address to pull bytes from
        :param length: Number of bytes to return
        :param default: Default byte value to use for anything not mapped in the disassembler.
            Defaults to throwing a NotExistError for invalid ranges.

        :return: Raw bytes from the opened file
        :raises NotExistError: If bytes within the given range does not exist.
        """

    @abc.abstractmethod
    def find_bytes(self, pattern: bytes, start: int = None, reverse: bool = False) -> int:
        """
        Search to find bytes for given pattern.

        NOTE: This just preforms simple byte matching. For something more complex
            like wild cards, I recommend using rugosa.re

        :param pattern: Bytes we are looking for
        :param start: Address to start the search (defaults to min or max address)
        :param reverse: Whether to search upwards instead of downwards.
        :return: The start address for the first found instance of given bytes or -1 if not found.
        """

    def get_word(self, addr: int) -> int:
        """
        Returns the two byte value at ``addr``.

        Must throw a NotExistError exception if the word cannot be retrieved.

        :param int addr: Address where the word is located
        :return: Two byte value
        :rtype: int
        """
        return int.from_bytes(self.get_bytes(addr, 2), "little")

    def get_dword(self, addr: int) -> int:
        """
        Returns the four byte value at ``addr``.

        Must throw a NotExistError exception if the dword cannot be retrieved.

        :param int addr: Address where the dword is located
        :return: Four byte value
        :rtype: int
        """
        return int.from_bytes(self.get_bytes(addr, 4), "little")

    def get_qword(self, addr: int) -> int:
        """
        Returns the eight byte value at ``addr``.

        Must throw a NotExistError exception if the qword cannot be retrieved.

        :param int addr: Address where the qword is located
        :return: Eight byte value
        :rtype: int
        """
        return int.from_bytes(self.get_bytes(addr, 8), "little")

    # TODO: determine if it would be better the other way around, where line's
    #   get_comment()/set_comment() call the flat api implementation.
    #   - OR should we only have comment stuff on flat and remove from line object?
    def get_comment(self, addr: int, comment_type=CommentType.eol) -> Optional[str]:
        """
        Obtains comment (of specific type) at line.

        :param addr int: Address of line to get comment from.
        :param comment_type: Type of comment to get (defaults to end of line comment)
        :returns: String containing comment or None if there is no comment.
        """
        return self.get_line(addr).get_comment(comment_type=comment_type)

    @abc.abstractmethod
    def get_function(self, addr: int) -> Function:
        """
        Returns a `Function` instance containing ``addr``.

        The :class:`~.Function` instance may be cached or a new instance created.

        :param int addr: Any address within the Function

        :return: A :class:`~.Function` instance containing the given address
        :raises NotExistError: If there is no function containing the given address.
        """

    def get_function_by_name(self, name: str, ignore_underscore: bool = True) -> Function:
        """
        Returns a `Function` instance with given name.

        :param str name: Name of function to obtain
        :param bool ignore_underscore: Whether to ignore leading or trailing underscores in function name.
            (Will return the first found function if enabled.)

        :return: A :class:`~.Function` instance containing the given address
        :raises NotExistError: If there is no function containing the given address.
        """
        for func in self.functions():
            func_name = func.name
            if ignore_underscore:
                func_name = func_name.strip("_")
            if func_name == name:
                return func
        raise NotExistError(f"Unable to find function with name: {name}")

    # TODO: Implement ability to provide operand for the assist?
    @abc.abstractmethod
    def get_function_signature(self, addr: int) -> FunctionSignature:
        """
        Returns a `FunctionSignature` instance for the function defined at ``addr``.
        Address can be the start of a function or address or an imported function.

        :param int addr: Any address
        :raises NotExistError: If there is no function signature at the given address.
        """

    def get_flowchart(self, addr: int) -> Flowchart:
        """
        Returns a `Flowchart` instance containing ``addr``.

        :param int addr: Any address within the Flowchart
        :raises NotExistError: If there is no flowchart containing the given address.
        """
        return self.get_function(addr).flowchart

    def get_basic_block(self, addr: int) -> BasicBlock:
        """
        Returns a `BasicBlock` instance containing ``addr``.

        :param int addr: Any address within the BasicBlock
        :raises NotExistError: If there is no basic block containing the given address.
        """
        return self.get_flowchart(addr).get_block(addr)

    def get_instruction(self, addr: int) -> Instruction:
        """
        Returns an Instruction object for the given any address contained within the
        instruction.

        :param int addr: Address of the instruction
        :return: Instruction object
        :raises NotExistError: If instruction isn't present at given address.
        """
        line = self.get_line(addr)
        instruction = line.instruction
        if instruction:
            return instruction
        else:
            raise NotExistError(f"Instruction not found at {hex(addr)}")

    @abc.abstractmethod
    def get_line(self, addr: int) -> Line:
        """
        Returns a Line object representing a defined item containing an instruction
        or data containing the given ``addr``

        If the address is not part of a line (instruction or data item), then
        it must throw a NotExistError exception.

        :param int addr: An address contained within the Line.
        """

    def get_line_address(self, addr: int) -> int:
        """
        Returns the address head of the 'line' containing ``addr``.  That is, the
        address of the start of the data item or instruction.

        For example, if there is an instruction at ``0x004010ef`` with 3 op codes and this
        method is called with ``get_line(0x004010f1)`` it must return ``0x004010ef``.

        :param int addr: Base address
        :return: Address head of the line which contains ``addr``
        :rtype: int
        """
        return self.get_line(addr).address

    def get_mnemonic(self, addr: int) -> str:
        """
        Returns the disassembler text of the indicated instruction's mnemonic.

        The output text is disassembler specific, but all implementations
        will likely be similar. However, each implementation should do its best
        to follow the official mnemonics as much as possible.

        :param int addr: Address of the instruction
        :return: String representation of the mnemonic
        :raises NotExistError: If instruction isn't present at given address.
        """
        instruction = self.get_instruction(addr)
        return instruction.mnemonic

    def get_name(self, addr: int) -> Optional[str]:
        """
        Gets the name of the specified location (if applicable)

        :param int addr: Address
        :returns: Name at address or None if address has no name.
        """
        # Get name for either the function or line
        func = None
        try:
            func = self.get_function(addr)
        except NotExistError:
            pass

        if func and func.start == addr:
            return func.name
        else:
            return self.get_line(addr).name

    def get_operand(self, addr: int, index: int) -> Operand:
        """
        Returns an Operand object based on the operand at a given address.
        """
        instruction = self.get_instruction(addr)
        operands = instruction.operands
        try:
            return operands[index]
        except IndexError:
            raise NotExistError(
                f"Instruction at {hex(instruction.address)} does not have an operand at index {index}"
            )

    def get_operand_type(self, addr: int, idx: int) -> OperandType:
        """
        Returns the type of the selected operand.

        The type is returned as an :class:`~OperandType` instance.

        If the address is invalid, this must throw a NotExistError exception.
        If the index is invalid, this must throw a IndexError exception.

        :param int addr: Address of the instruction
        :param int idx: Index of the operand
        :return: Operand type
        """
        operand = self.get_operand(addr, idx)
        return operand.type

    def get_operand_value(self, addr: int, idx: int) -> Optional[OperandValue]:
        """
        The number used in the operand.
        This function returns an immediate number used in the operand
        based on the type of operand.

        :return: OperandValue object
        """
        operand = self.get_operand(addr, idx)
        return operand.value

    @abc.abstractmethod
    def get_data_type(self, name: str) -> DataType:
        """
        Obtain the data type from the given name.
        """

    @abc.abstractmethod
    def get_register(self, name: str) -> Register:
        """
        Obtains a register from given name.

        :param str name: Name of register.
        :return: Register object for given register.
        :raises NotExistError: If register of given name is not a valid processor register.
        """

    @abc.abstractmethod
    def get_segment(self, addr_or_name: Union[int, str]) -> Segment:
        """
        Returns an open memory segment object that contains the given address or has the given name.
        (If providing a name and there are multiple segments with the given name,
        it returns the first one.)

        :param addr_or_name: Either an address contained in the segment or name of the segment.
        :return: Segment object.
        :rtype: Segment
        :raises NotExistError: If a segment doesn't exist for the given input.
        """

    @abc.abstractmethod
    def create_segment(self, name: str, start: int, size: int) -> Segment:
        """
        Creates and returns a memory segment object that starts at given address.

        :param name: Name of segment
        :param start: Start address of segment
        :param size: Size of segment
        :return: Segment object created
        :raises ValueError: If segment couldn't be created.
        """

    @property
    @abc.abstractmethod
    def segments(self) -> Iterable[Segment]:
        """
        Iterates (initialized) segments found in the program.
        """

    @abc.abstractmethod
    def get_string_bytes(self, addr: int, length: int = None, bit_width: int = None) -> bytes:
        """
        Returns the raw bytes for a string at the represented address.
        (This effectively gets bytes until it encounters the null terminator)

        WARNING: This function should return just the raw bytes.
        User is responsible for decoding data as appropriate.

        :param int addr: Start address to pull string data.
        :param int length: Max length of string data.
            (May be omitted to allow disassembler to determine appropriate value.)
        :param int bit_width: Number of bits which represent a code point.
            (May be omitted to allow disassembler to determine appropriate value.)
            (Must be 8, 16, or 32 if provided.)
        :return: Raw bytes containing encoded string.
        """

    @abc.abstractmethod
    def strings(self, min_length=3) -> Iterable[String]:
        """
        Iterates strings found in the disassembly.

        :param min_length: Minimum length of string to have the found string count.

        :yields: String object
        """

    def line_addresses(self, start: int = None, end: int = None, reverse=False) -> Iterable[int]:
        """
        Returns a list of all of the address heads between ``start`` and ``end``.

        The range of address heads must be ``[start, end)``. That is,
        if there is an address head at ``start``, it is included in the list, but if
        an address head is at ``end``, it is not included.

        If there are no address heads in the given range or the range is invalid
        (e.g. ``start >= end``) an empty list must be returned.

        :param int start: Start address (included)
        :param int end: End address (excluded)
        :return: Iterates start addresses.
        """
        for line in self.lines(start=start, end=end, reverse=reverse):
            yield line.address

    def lines(self, start: int = None, end: int = None, reverse=False) -> Iterable[Line]:
        """
        Iterates the Line objects found with the given range of addresses.
        If addresses are not found, the address among the full program is iterated.

        By default we take advantage of the .prev and .next properties of Line
        to implement this functionality. However, this may be overwritten if optimization
        is necessary.

        :param start: Start address (included)
        :param end: End address (excluded)
        :param reverse: Whether to iterate lines in reverse.
        """
        if start is None:
            start = self.min_address if not reverse else self.max_address
        if end is None:
            end = self.max_address if not reverse else self.min_address

        # TODO: Automatically detect reverse if start > end?
        if reverse:
            if end > start:
                raise ValueError(f"Start address {hex(start)} must be greater than end address {hex(end)}")
            line = self.get_line(start)
            while True:
                if line.address <= end:
                    break
                yield line
                line = line.prev
                if not line:
                    break
        else:
            if end < start:
                raise ValueError(f"Start address {hex(start)} must be less than end address {hex(end)}")
            line = self.get_line(start)
            while True:
                if line.address >= end:
                    break
                yield line
                line = line.next
                if not line:
                    break

    def instructions(self, start: int = None, end: int = None, reverse=False) -> Iterable[Instruction]:
        """
        Iterates the Instruction objects found with the given range of addresses.
        If addresses are not found, the address among the full program is iterated.
        """
        for line in self.lines(start=start, end=end, reverse=reverse):
            insn = line.instruction
            if insn:
                yield insn

    @property
    @abc.abstractmethod
    def max_address(self) -> int:
        """
        Maximum address in program.
        """

    @property
    @abc.abstractmethod
    def min_address(self) -> int:
        """
        Minimum address in program.
        """

    @property
    @abc.abstractmethod
    def base_address(self) -> int:
        """
        Image base address.
        """

    def next_line_address(self, addr: int) -> int:
        """
        Returns the address head of the 'line', that is, instruction or data,
        following the specified address.

        If a following line does not exist (e.g. the base address is the last
        address of the image, or after it) then this must throw a NotExistError
        exception.

        :param int addr: Base address
        :return: Address head of following line
        :rtype: int
        """
        return self.get_line(addr).next.address

    def prev_line_address(self, addr: int) -> int:
        """
        Returns the address head of the 'line', that is, instruction or data,
        preceding the specified address.

        If a preceding line does not exist (e.g. the base address is the first
        address of the image, or before it) then this must throw a NotExistError
        exception.

        :param int addr: Base address
        :return: Address head of previous line
        :rtype: int
        """
        return self.get_line(addr).prev.address

    @abc.abstractmethod
    def open_memory(self, start: int, end: int) -> Memory:
        """
        Opens and returns a file-like object backed by the given memory window.

        :param start: Starting address of the memory window.
        :param end: Ending address of the memory window (non-inclusive).
        :return: Memory object.
        """

    @abc.abstractmethod
    def references_from(self, addr: int) -> Iterable[Reference]:
        """
        Iterates cross references from the specified address.

        :param int addr: Address to get references from
        :yield: `Reference` objects.
        """

    @abc.abstractmethod
    def references_to(self, addr: int) -> Iterable[Reference]:
        """
        Iterates cross references to the specified address.

        :param int addr: Address to get references to
        :yield: `Reference` objects.
        """

    @abc.abstractmethod
    def create_reference(self, from_address: int, to_address: int, ref_type: ReferenceType) -> Reference:
        """
        Creates a cross reference.

        :param from_address: source address
        :param to_address: destination address
        :param ref_type: Type of reference
        :return: Reference object created
        :raises ValueError: If cross reference failed to create.
        """

    def get_variable(self, addr: int) -> GlobalVariable:
        """
        Obtains the global variable containing the given address.

        :param int addr: An address contained in the variable.
        :raises NotExistError: If a variable doesn't exist for the given address.
        """

    def set_comment(self, addr: int, comment: Optional[str], comment_type=CommentType.eol):
        """
        Sets comment (of specific type) at line.

        :param addr int: Address of line to set comment.
        :param str comment: comment string to set or None to reset comment.s
        :param comment_type: Type of comment to set (default to end of line comment)
        """
        self.get_line(addr).set_comment(comment, comment_type=comment_type)

    # TODO: Handle duplicate names, by adding a prefix, ala IDA, but do it locally.
    def set_name(self, addr: int, name: Optional[str]):
        """
        Sets the name of specified location.

        The name text may be a unicode or bytes string. Any manipulation
        of the name to meet disassembler-specific restrictions are
        done by the implementation. This means what is given as ``name``
        may not be exactly what is saved by the disassembler.

        How this is precisely implemented is disassembler-specific, however,
        in general if ``addr`` is the starting address of a function, then
        the function at that location should have its name changed to ``name``.

        If the name was not successfully set, then this must throw an exception

        :param int addr: Address to rename
        :param name: The new name of the address or None/"" to reset the name.
        """
        # Set name for either the function or line
        func = None
        try:
            func = self.get_function(addr)
        except NotExistError:
            pass

        if func and func.start == addr:
            func.name = name
        else:
            self.get_line(addr).name = name

    @property
    @abc.abstractmethod
    def imports(self) -> Iterable[Import]:
        """
        The imports within the binary.
        """

    def get_import(self, name: str) -> Import:
        """
        Gets import symbol by name.
        :param name: Name of import function
        :return: Import symbol
        :raises NotExistError: If import by the given name doesn't exist.
        """
        for import_ in self.imports:
            if import_.name == name:
                return import_
        raise NotExistError(f"Import with name '{name}' doesn't exist.")

    @property
    @abc.abstractmethod
    def exports(self) -> Iterable[Export]:
        """
        The exports within the binary.
        """

    def get_export(self, name: str) -> Export:
        """
        Gets export symbol by name.
        :param name: Name of import function
        :return: Export symbol
        :raises NotExistError: If export by the given name doesn't exist.
        """
        for export in self.exports:
            if export.name == name:
                return export
        raise NotExistError(f"Export with name '{name}' doesn't exist.")




from __future__ import annotations
from typing import TYPE_CHECKING, Union

from dragodis.interface.variable import StackVariable, GlobalVariable
from dragodis.ghidra.data_type import GhidraDataType

if TYPE_CHECKING:
    from dragodis.ghidra.flat import GhidraFlatAPI
    import ghidra


class GhidraGlobalVariable(GlobalVariable):

    def __init__(self, ghidra: GhidraFlatAPI, data: "ghidra.program.model.listing.Data"):
        self._ghidra = ghidra
        self._data = data

    def __eq__(self, other) -> bool:
        return isinstance(other, GhidraGlobalVariable) and self.address == other.address

    @property
    def address(self) -> int:
        return self._data.getAddress().getOffset()

    @property
    def name(self) -> str:
        return self._data.getPathName()

    @name.setter
    def name(self, new_name: str):
        from ghidra.program.model.symbol.SourceType import SourceType
        symbol = self._ghidra._flatapi.getSymbolAt(self._data.getAddress())
        symbol.setName(new_name, SourceType.USER_DEFINED)

    @property
    def data_type(self) -> GhidraDataType:
        return GhidraDataType(self._data.getDataType())

    @property
    def size(self) -> int:
        return self._data.getLength()


class GhidraStackVariable(StackVariable):

    def __init__(self, ghidra: GhidraFlatAPI, variable: "ghidra.program.model.listing.Variable"):
        if not variable.isStackVariable():
            raise ValueError(f"{variable} must be a stack variable")
        self._ghidra = ghidra
        self._variable = variable

    def __eq__(self, other) -> bool:
        return isinstance(other, GhidraStackVariable) and self._variable.isEquivalent(other._variable)

    @property
    def stack_offset(self) -> int:
        return self._variable.getStackOffset()

    @property
    def name(self) -> str:
        return self._variable.getName()

    @name.setter
    def name(self, new_name: str):
        from ghidra.program.model.symbol.SourceType import SourceType
        self._variable.setName(new_name, SourceType.USER_DEFINED)

    @property
    def data_type(self) -> GhidraDataType:
        return GhidraDataType(self._variable.getDataType())

    # @data_type.setter
    # def data_type(self, data_type: DataType):
    #     self._variable.setDataType(data_type)

    @property
    def size(self) -> int:
        return self._variable.getLength()


GhidraVariable = Union[GhidraGlobalVariable, GhidraStackVariable]




from __future__ import annotations
import abc
import io
import os


class Memory(io.BufferedIOBase, metaclass=abc.ABCMeta):
    """
    Interface for reading and writing a window of memory within the program
    as a file-like object.
    """

    def __init__(self, start: int, end: int):
        """
        :param start: The start address of the memory window.
        :param end: The end address of the memory window (non-inclusive).
        """
        if start > end:
            raise ValueError(f"Start address must be less than end address.")
        self.start = start
        self.end = end
        self._offset = 0

    def __repr__(self) -> str:
        return f"<Memory 0x{self.start:08x} - 0x{self.end:08x}>"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        """
        Closes the memory buffer window if applicable.
        NOTE: Ensure you call super().close() if overriding this function.
        """
        # We don't need to do anything here, but want to make sure we provide a
        # docstring to warn users.
        super().close()

    @property
    def data(self) -> bytes:
        """
        Raw bytes that make up the underlying data within this memory window.
        WARNING: This can consume a lot of memory.
        Defaults to consuming all of the data stream.
        """
        offset = self.tell()
        try:
            self.seek(0)
            return self.read()
        finally:
            self.seek(offset)

    def flush(self):
        """
        Flush the write buffers of the memory window if applicable.
        """
        super().flush()

    @abc.abstractmethod
    def read(self, size: int = None) -> bytes:
        """
        Reads and returns size number of bytes from window.
        Reads in all data if size is omitted.

        :param size: Number of bytes to read.
        :return: Read bytes.
        """

    def readable(self) -> bool:
        return True

    @abc.abstractmethod
    def reset(self, size: int = None) -> int:
        """
        Resets given number of bytes back to original unpatched binary.

        :param size: Number of bytes to unpatch.
        :return: Number of bytes successfully unpatched.
        """

    @abc.abstractmethod
    def write(self, data: bytes) -> int:
        """
        Writes given data to memory window.

        :param data: Bytes to write to memory window.
        :return: Number of bytes successfully written.
        :raises IOError: If attempted to write to an uninitialized section.
        """

    def writable(self) -> bool:
        return True

    def seek(self, offset: int, whence: int = os.SEEK_SET) -> int:
        """
        Sets the current offset, relative to whence flag.

        :param offset: Offset to set to.
        :param whence: Flag to indicate positioning.
        :return: The new absolute position.
        """
        if whence == os.SEEK_SET:
            self._offset = offset
        elif whence == os.SEEK_CUR:
            self._offset += offset
        elif whence == os.SEEK_END:
            self._offset = (self.end - self.start) + offset
        else:
            raise ValueError(f"Invalid whence: {whence}")

        # Fix up offset if out of bounds.
        if self._offset < 0:
            self._offset = 0
        elif self._offset > self.end:
            self._offset = self.end

        return self._offset

    def seek_address(self, addr: int) -> int:
        """
        Sets memory window to given address.

        :param addr: Address within memory window to seek to.
        :return: The new absolute position (as an address).
        :raises IOError: If address is not within segment.
        """
        if addr < self.start or addr > self.end:
            raise IOError(f"Invalid seek address: 0x{addr:08x}")
        return self.seek(addr - self.start) + self.start

    def seekable(self) -> bool:
        return True

    def tell(self) -> int:
        """
        Returns the current offset within the memory window.
        """
        return self._offset

    def tell_address(self) -> int:
        """
        Returns the virtual address representation of the current offset in the memory window.
        """
        return self.start + self.tell()



"""
Helper utilities for Ghidra
"""
from typing import TYPE_CHECKING

from dragodis.interface import FlowType

if TYPE_CHECKING:
    import ghidra


def iterate(java_iterator):
    """
    Sometimes jpype doesn't pick up some of Ghidra's Iterator objects in order
    to automatically convert it to a python iterator.
    Therefore, this can be used to manually iterate a given java iterator as a python generator.
    """
    while java_iterator.hasNext():
        yield java_iterator.next()


def convert_flow_type(flow_type: "ghidra.program.model.symbol.FlowType") -> FlowType:
    """
    Converts a Ghidra FlowType to a Dragodis FlowType

    :raises TypeError: Encountered unknown/unsupported flow type.
    """
    if flow_type.isCall():
        return FlowType.call
    elif flow_type.isTerminal():
        return FlowType.terminal
    elif flow_type.isJump():
        if flow_type.isConditional():
            return FlowType.conditional_jump
        elif flow_type.isUnConditional():
            return FlowType.unconditional_jump
    elif flow_type.isFallthrough():
        return FlowType.fall_through

    raise TypeError(f"Unknown flow type: {flow_type}")




from __future__ import annotations
from typing import TYPE_CHECKING

from dragodis.interface.string import String

if TYPE_CHECKING:
    from dragodis.ghidra.flat import GhidraFlatAPI
    import ghidra


class GhidraString(String):

    def __init__(self, ghidra: GhidraFlatAPI, data_item: "ghidra.program.model.listing.Data"):
        self._ghidra = ghidra
        self._data_item = data_item

    @property
    def value(self) -> str:
        return str(self._data_item.value)

    @property
    def data(self) -> bytes:
        return self._ghidra.get_bytes(self.address, self._data_item.getLength()).rstrip(b"\x00")

    @property
    def address(self) -> int:
        return self._data_item.getAddress().getOffset()




from __future__ import annotations
import abc
from typing import Optional, TYPE_CHECKING, Any, Iterable

from dragodis.interface.types import CommentType, LineType


if TYPE_CHECKING:
    from dragodis.interface import Instruction, FlatAPI, Reference


class Line(metaclass=abc.ABCMeta):
    """
    Interface for a defined item (instruction or data).
    (This represents a "line" as you see in the disassembly view.)
    """

    def __init__(self, api: FlatAPI):
        self._api = api

    def __len__(self):
        """Here for convenience"""
        return self.size

    def __str__(self):
        return f"0x{self.address:08x}: {self.value}"

    def __repr__(self):
        return f"<Line 0x{self.address:08x} - {self.value!r}>"

    @property
    @abc.abstractmethod
    def address(self) -> int:
        """
        Starting address of the line
        """

    @property
    @abc.abstractmethod
    def data(self) -> bytes:
        """
        Raw bytes that make up this line.
        """

    @data.setter
    @abc.abstractmethod
    def data(self, new_data: bytes):
        """
        Sets the data that represents the line.
        This patches the underlying disassembler with the given data.

        WARNING: The new_data provided should be of the same size as the original.
        """

    @abc.abstractmethod
    def get_comment(self, comment_type=CommentType.eol) -> Optional[str]:
        """
        Obtains comment (of specific type) at line.

        :param comment_type: Type of comment to get (defaults to end of line comment)
        :returns: String containing comment or None if there is no comment.
        """

    @property
    def instruction(self) -> Optional[Instruction]:
        """
        The Instruction object for the given line (if line is an instruction)
        Returns None if not an instruction.
        """
        if self.is_code:
            return self.value

    @property
    def is_code(self) -> bool:
        """
        Convenience property for determining if the line is an instruction.
        """
        return self.type == LineType.code

    @property
    def is_float(self) -> bool:
        """
        Convenience property for determing if the line is a float type.
        """
        return self.type in (LineType.float, LineType.double)

    @property
    def is_integer(self) -> bool:
        """
        Convenience property for determine if the line is an integer.
        """
        return self.type in (LineType.word, LineType.dword, LineType.qword, LineType.oword)

    @property
    def is_loaded(self) -> bool:
        """
        Convenience property for determining if the line is loaded.
        """
        return self.type != LineType.unloaded

    @property
    def is_string(self) -> bool:
        """
        Convenience property for determingin if the line is a string (with any encoding).
        """
        return self.type in (LineType.string, LineType.string16, LineType.string32)

    @property
    @abc.abstractmethod
    def name(self) -> Optional[str]:
        """
        The current name or label on the line.
        Returns None if line doesn't have a name.
        """

    # TODO: Since the name actual set by the disassembler may be different than
    #   what was passed, should we make this a function "set_name()" or "rename()"
    #   instead?
    # TODO: Should we have our own standardization for modifying duplicate names so
    #   they are consistent across disassemblers?
    @name.setter
    @abc.abstractmethod
    def name(self, value: str):
        """
        Sets the name of the line.
        """

    @property
    @abc.abstractmethod
    def next(self) -> Optional["Line"]:
        """
        The next line (if it exists).
        """

    @property
    @abc.abstractmethod
    def prev(self) -> Optional["Line"]:
        """
        The previous line (if it exists).
        """

    @abc.abstractmethod
    def set_comment(self, comment: Optional[str], comment_type=CommentType.eol):
        """
        Sets comment (of specific type) at line.

        :param str comment: comment string to set.
            (If value is an empty string or None, existing comments are cleared)
        :param comment_type: Type of comment to set (default to end of line comment)
        """

    @property
    @abc.abstractmethod
    def size(self) -> int:
        """
        Number of bytes of data that make up this line.
        """

    @property
    @abc.abstractmethod
    def type(self) -> LineType:
        """
        The type of line
        """

    @type.setter
    @abc.abstractmethod
    def type(self, new_type: LineType):
        """
        Sets the type that represents the line.
        This patches the underlying disassembler to the given type.

        WARNING: Changing this to a type that is larger and smaller will cause
        changes to the .next value, be sure to clear it's cache in your implementation.
        """

    @abc.abstractmethod
    def undefine(self):
        """
        Undefines the line, clearing the line of any type annotation.

        WARNING: Be sure to clear the property cache after clearing.
        TODO: We should have a utility for clearing the caches of all the properties.
        """

    @property
    @abc.abstractmethod
    def value(self) -> Any:
        """
        A value that represents the line.
        The return type is dependent upon the type of line.

        e.g.
            code -> Instruction
            dword -> int
            word -> int
            float -> float
            double -> float
            string -> str
            align -> bytes
            unknown -> bytes
        """

    @value.setter
    @abc.abstractmethod
    def value(self, new_value: Any):
        """
        Sets the value that represents the line.
        This patches the underlying disassembler with the given value.
        If the data type changes, the type will be patched too.
        """

    @property
    def references_from(self) -> Iterable[Reference]:
        """
        Iterates cross references from the specified address.

        :param int addr: Address to get references from
        :yield: `Reference` objects.
        """
        yield from self._api.references_from(self.address)

    @property
    def references_to(self) -> Iterable[Reference]:
        """
        Iterates cross references to the specified address.

        :param int addr: Address to get references to
        :yield: `Reference` objects.
        """
        yield from self._api.references_to(self.address)




from __future__ import annotations
from typing import List, Union, TYPE_CHECKING

from dragodis.ghidra.data_type import GhidraDataType
from dragodis.ghidra.function_argument_location import (
    GhidraArgumentLocation, GhidraStackLocation,
    GhidraRegisterLocation, GhidraRegisterPairLocation, GhidraRelativeRegisterLocation, GhidraStaticLocation
)
from dragodis.interface.function_signature import FunctionSignature, FunctionParameter
from dragodis.exceptions import NotExistError

if TYPE_CHECKING:
    import ghidra
    from dragodis.ghidra.flat import GhidraFlatAPI


class GhidraFunctionSignature(FunctionSignature):

    def __init__(self, ghidra: GhidraFlatAPI, function: "ghidra.program.model.listing.Function"):
        self._ghidra = ghidra
        self._function = function
        if not self._is_fully_defined():
            self._complete_function_signature()

    @property
    def name(self) -> str:
        return self._function.getSignature().getName()

    @property
    def declaration(self) -> str:
        # Including calling convention to be consistent with IDA.
        return self._function.getSignature().getPrototypeString(True)

    @property
    def calling_convention(self) -> str:
        cc = self._function.getCallingConvention()
        if cc is None:
            return ""
        return cc.getName()

    @calling_convention.setter
    def calling_convention(self, name: str):
        if not name.startswith("__"):
            name = f"__{name}"
        from ghidra.util.exception import InvalidInputException
        try:
            self._function.setCallingConvention(name.lower())
        except InvalidInputException as e:
            raise ValueError(e)

    @property
    def return_type(self) -> GhidraDataType:
        return GhidraDataType(self._function.getReturnType())

    @return_type.setter
    def return_type(self, data_type: Union[GhidraDataType, str]):
        from ghidra.program.model.symbol import SourceType
        if isinstance(data_type, str):
            data_type = self._ghidra.get_data_type(data_type)
        self._function.setReturnType(data_type._data_type, SourceType.USER_DEFINED)

    @property
    def parameters(self) -> List[FunctionParameter]:
        return [
            GhidraFunctionParameter(self, parameter)
            for parameter in self._function.getParameters()
        ]

    def _generate_parameter(self, data_type: str) -> "ghidra.program.model.listing.Parameter":
        """
        Generates a Ghidra Parameter objects from declaration.
        """
        from ghidra.program.model.listing import ParameterImpl
        from ghidra.program.model.symbol import SourceType

        data_type = self._ghidra.get_data_type(data_type)._data_type
        return ParameterImpl(
            "",
            data_type,
            self._function.getEntryPoint(),
            self._ghidra._program,
            SourceType.USER_DEFINED
        )

    def replace_parameters(self, data_types: List[str]):
        parameters = [self._generate_parameter(data_type) for data_type in data_types]
        self._set_parameters(parameters)

    def _set_parameters(self, parameters: List["ghidra.program.model.listing.Parameter"]):
        from ghidra.program.model.listing.Function import FunctionUpdateType
        from ghidra.program.model.symbol import SourceType
        self._function.replaceParameters(
            FunctionUpdateType.DYNAMIC_STORAGE_ALL_PARAMS,
            False,
            SourceType.USER_DEFINED,
            parameters,
        )

    def add_parameter(self, data_type: str) -> GhidraFunctionParameter:
        from ghidra.program.model.symbol import SourceType
        new_param = self._generate_parameter(data_type)
        new_param = self._function.addParameter(new_param, SourceType.USER_DEFINED)
        return GhidraFunctionParameter(self, new_param)

    def remove_parameter(self, ordinal: int):
        num_parameters = self._function.getParameterCount()
        if ordinal < 0:
            ordinal += num_parameters
        if ordinal not in range(num_parameters):
            raise NotExistError(f"Invalid ordinal for parameter deletion: {ordinal}")
        self._function.removeParameter(ordinal)

    def insert_parameter(self, ordinal: int, data_type: str) -> GhidraFunctionParameter:
        num_parameters = self._function.getParameterCount()
        if ordinal < 0:
            ordinal += num_parameters
        if ordinal not in range(num_parameters):
            raise ValueError(f"Invalid ordinal for parameter insertion: {ordinal}")
        from ghidra.program.model.symbol import SourceType
        new_param = self._generate_parameter(data_type)
        new_param = self._function.insertParameter(ordinal, new_param, SourceType.USER_DEFINED)
        return GhidraFunctionParameter(self, new_param)

    def _is_fully_defined(self):
        """
        Checks if the function signature has already been set
        """
        from ghidra.program.model.symbol import SourceType
        source = self._function.getSignatureSource()
        return source.isHigherPriorityThan(SourceType.DEFAULT)

    def _complete_function_signature(self):
        """
        Decompiles the function to detect and add any missing arguments.
        Function arguments are usually missing when a custom calling convention
        is used or when the function signature has not been committed.
        """
        decomp = self._decompile()
        if decomp is None or not decomp.decompileCompleted():
            return

        from ghidra.program.model.data import Undefined
        from ghidra.program.model.listing import Function, ParameterImpl
        from ghidra.program.model.symbol import SourceType

        unresolved_params = [p for p in self._get_unresolved_parameters(decomp)]

        program = self._ghidra._program

        params = []

        # collect the existing parameters
        high_proto = decomp.getHighFunction().getFunctionPrototype()
        for i in range(high_proto.getNumParams()):
            sym = high_proto.getParam(i)
            storage = sym.getStorage()
            dt = Undefined.getUndefinedDataType(sym.getSize())
            param = ParameterImpl(None, dt, storage, program)
            params.append(GhidraFunctionParameter(self, param))

        if not unresolved_params:
            # no custom storage
            # set the parameters and return
            self._set_parameters([param._parameter for param in params])
            return

        # add the detected missing parameters to the signature
        for unresolved_param in unresolved_params:
            storage = unresolved_param.getStorage()
            dt = Undefined.getUndefinedDataType(unresolved_param.getSize())
            param = ParameterImpl(None, dt, storage, program)
            params.append(GhidraFunctionParameter(self, param))

        # sort them so parameter ordering is consistent
        params.sort(key=lambda param: param.location)

        # fetch the original Ghidra parameter types
        params = [param._parameter for param in params]

        CUSTOM_STORAGE = Function.FunctionUpdateType.CUSTOM_STORAGE
        self._function.setCustomVariableStorage(True)
        self._function.replaceParameters(CUSTOM_STORAGE, True, SourceType.USER_DEFINED, *params)

    def _decompile(self):
        ghidra = self._ghidra
        return ghidra._decomp_api.decompiler.decompileFunction(self._function, 0, ghidra._monitor)

    def _get_unresolved_parameters(self, decomp):
        high_function = decomp.getHighFunction()
        if not high_function:
            return
        for sym in high_function.getLocalSymbolMap().getSymbols():
            name = sym.getName()

            # unaff_ may be problematic and may indicate an unconventional output somewhere
            if not (name.startswith("in_") or name.startswith("unaff_")):
                continue
            if name in ("in_FS_OFFSET", "in_GS_OFFSET"):
                # these are pseudo registers and are never parameters
                continue
            yield sym


# TODO: Parameter should have an option to get a "Variable" object which represents the location?
class GhidraFunctionParameter(FunctionParameter):

    def __init__(self, signature: GhidraFunctionSignature, parameter: "ghidra.program.model.listing.Parameter"):
        super().__init__(signature)
        self.signature: GhidraFunctionSignature
        self._ghidra = signature._ghidra
        self._parameter = parameter
        self._function = signature._function

    def __eq__(self, other):
        return (
            isinstance(other, GhidraFunctionParameter)
            and self._parameter.isEquivalent(other._parameter)
        )

    @property
    def name(self) -> str:
        return self._parameter.getName()

    @name.setter
    def name(self, new_name: str):
        from ghidra.program.model.symbol import SourceType
        self._parameter.setName(new_name, SourceType.USER_DEFINED)

    @property
    def ordinal(self) -> int:
        return self._parameter.getOrdinal()

    @property
    def size(self) -> int:
        return self._parameter.getLength()

    @property
    def data_type(self) -> GhidraDataType:
        return GhidraDataType(self._parameter.getDataType())

    @data_type.setter
    def data_type(self, new_type: Union[str, GhidraDataType]):
        from ghidra.program.model.symbol import SourceType
        if isinstance(new_type, str):
            new_type = self._ghidra.get_data_type(new_type)
        self._parameter.setDataType(new_type._data_type, SourceType.USER_DEFINED)

    @property
    def declaration(self) -> str:
        return str(self._parameter)

    @property
    def location(self) -> GhidraArgumentLocation:
        storage = self._parameter.getVariableStorage()

        if storage.isStackStorage():
            return GhidraStackLocation(self._function, storage)
        elif storage.isRegisterStorage():
            return GhidraRegisterLocation(self._function, storage)
        elif storage.isCompoundStorage() and len(storage.getRegisters()) == 2:
            return GhidraRegisterPairLocation(self._function, storage)
        # TODO: I'm guessing on this one, need to find a sample to test this.
        elif (
            storage.isCompoundStorage()
            and storage.getVarnodeCount() == 2
            and storage.getFirstVarnode().isRegister()
            and storage.getLastVarnode().isConstant()
        ):
            return GhidraRelativeRegisterLocation(self._function, storage)
        elif storage.isMemoryStorage():
            return GhidraStaticLocation(self._function, storage)
        else:
            raise ValueError(f"Unsupported argument location type: {storage}")




from __future__ import annotations

from functools import lru_cache

from .. import utils

from typing import Iterable, Union, Optional

from dragodis.interface.flat import FlatAPI
from dragodis.exceptions import NotExistError
from .data_type import IDADataType
from .disassembler import IDADisassembler, IDARemoteDisassembler, IDALocalDisassembler
from .function_signature import IDAFunctionSignature
from .line import IDALine
from .memory import IDAMemory, CachedMemory
from .function import IDAFunction
from .operand_value import IDARegister
from .reference import IDAReference
from .segment import IDASegment
from .string import IDAString
from .symbol import IDAImport, IDAExport
from .variable import IDAGlobalVariable
from ..interface import ReferenceType


cache = lru_cache(maxsize=1024)


class IDAFlatAPI(FlatAPI, IDADisassembler):

    @property
    def _cached_memory(self):
        return CachedMemory(self)

    def _bytes_loaded(self, addr: int, num_bytes: int) -> bool:
        return self._ida_helpers.is_loaded(addr, num_bytes)

    @property
    def processor_name(self) -> str:
        proc = self._ida_ida.inf_get_procname()
        # Switching "metapc" to "x86" to match Ghidra.
        if proc == "metapc":
            return "x86"
        return proc

    @property
    def compiler_name(self) -> str:
        cc_id = self._ida_ida.inf_get_cc_id()
        return self._ida_typeinf.get_compiler_name(cc_id)

    @property
    def bit_size(self) -> int:
        # IDA 7.6 adds ida_ida.inf_get_app_bitness()
        if self._idaapi.IDA_SDK_VERSION >= 760:
            return self._ida_ida.inf_get_app_bitness()

        if self._ida_ida.inf_is_64bit():
            return 64
        elif self._ida_ida.inf_is_32bit():
            return 32
        else:
            return 16

    @property
    def is_big_endian(self) -> bool:
        return self._ida_ida.inf_is_be()

    @property
    def entry_point(self) -> Optional[int]:
        address = self._ida_ida.inf_get_start_ip()
        if address != self._BADADDR:
            return address

    @cache
    def get_virtual_address(self, file_offset: int) -> int:
        addr = self._ida_loader.get_fileregion_ea(file_offset)
        if addr == self._idc.BADADDR:
            raise NotExistError(f"Cannot get linear address for file offset: {hex(file_offset)}")
        return addr

    @cache
    def get_file_offset(self, addr: int) -> int:
        file_offset = self._ida_loader.get_fileregion_offset(addr)
        if file_offset == -1:
            raise NotExistError(f"Cannot get file offset for address: {hex(addr)}")
        return file_offset

    def functions(self, start=None, end=None) -> Iterable[IDAFunction]:
        # TODO: use remove_eval to optimize this?
        #   - Create mechanism to chunk iterators?
        #   - Reimplement idautils.Functions here?
        for ea in self._idautils.Functions(start=start, end=end):
            # IDA will include the function if started in the middle of it.
            # Ignore this function to stay consistent with Ghidra.
            if start and ea < start:
                continue
            yield self.get_function(ea)

    # TODO: make a Memory object?

    def get_byte(self, addr: int) -> int:
        if not self._bytes_loaded(addr, 1):
            raise NotExistError(f"Cannot get byte at {hex(addr)}")
        return self._ida_bytes.get_wide_byte(addr)

    def get_bytes(self, addr: int, length: int, default: int = None) -> bytes:
        if default is None and not self._ida_helpers.is_loaded(addr, length):
            raise NotExistError(
                f"Unable to obtain {length} bytes from 0x{addr:08X}: "
                f"Address range not fully loaded."
            )
        return self._ida_helpers.get_bytes(addr, length, default=default or 0)
        # FIXME: Disabling use of cached memory since we aren't invalidating caches properly.
        # # If all bytes aren't available but a default was provided, get bytes
        # # one at a time, replacing invalid bytes with the default.
        # if default:
        #     default = bytes([default])
        # return self._cached_memory.get(addr, length, fill_pattern=default)

    def find_bytes(self, pattern: bytes, start: int = None, reverse=False) -> int:
        # Convert bytes pattern into hex separated by space format that IDA likes.
        pattern = " ".join(format(byte, "02X") for byte in pattern)
        if reverse:
            flag = self._ida_search.SEARCH_UP
            if start is None:
                start = self.max_address
            end = self.min_address
        else:
            flag = self._ida_search.SEARCH_DOWN
            if start is None:
                start = self.min_address
            end = self.max_address

        found = self._ida_search.find_binary(start, end, pattern, 16, flag)
        if found == self._BADADDR:
            return -1
        else:
            return found

    def get_word(self, addr: int) -> int:
        if not self._bytes_loaded(addr, 2):
            raise NotExistError(f"Cannot get word at {hex(addr)}")
        return self._ida_bytes.get_wide_word(addr)

    def get_dword(self, addr: int) -> int:
        if not self._bytes_loaded(addr, 4):
            raise NotExistError(f"Cannot get dword at {hex(addr)}")
        return self._ida_bytes.get_wide_dword(addr)

    def get_qword(self, addr: int) -> int:
        if not self._bytes_loaded(addr, 8):
            raise NotExistError(f"Cannot get qword at {hex(addr)}")
        return self._ida_bytes.get_qword(addr)

    def get_function(self, addr: int) -> IDAFunction:
        func_t = self._ida_funcs.get_func(addr)
        if not func_t:
            raise NotExistError(f"Function does not exist at {hex(addr)}")
        return IDAFunction(self, func_t)

    def get_function_by_name(self, name: str, ignore_underscore: bool = True) -> IDAFunction:
        # Using ida_helpers instead of default implementation to improve performance.
        func_t = self._ida_helpers.get_function_by_name(name, ignore_underscore=ignore_underscore)
        if func_t is None:
            raise NotExistError(f"Unable to find function with name: {name}")
        return IDAFunction(self, func_t)

    # TODO: Add support for providing an operand to help get a better function signature type.
    @cache
    def get_function_signature(self, addr: int) -> IDAFunctionSignature:
        # Constructor will raise a NotExistError if we can't make a function signature.
        return IDAFunctionSignature(self, addr)

    def get_line(self, addr: int) -> IDALine:
        return IDALine(self, addr)

    def get_register(self, name: str) -> IDARegister:
        reg_info = self._ida_idp.reg_info_t()
        success = self._ida_idp.parse_reg_name(reg_info, name)
        if not success:
            raise NotExistError(f"Invalid register name: {name}")
        return IDARegister(self, reg_info.reg, reg_info.size)

    @cache
    def get_segment(self, addr_or_name: Union[int, str]) -> IDASegment:
        if isinstance(addr_or_name, str):
            name = addr_or_name
            segment_t = self._ida_segment.get_segm_by_name(name)
            if not segment_t:
                raise NotExistError(f"Could not find segment with name: {name}")
        elif isinstance(addr_or_name, int):
            addr = addr_or_name
            segment_t = self._ida_segment.getseg(addr)
            if not segment_t:
                raise NotExistError(f"Could not find segment containing address: 0x{addr:08x}")
        else:
            raise ValueError(f"Invalid input: {addr_or_name!r}")

        return IDASegment(self, segment_t)

    def create_segment(self, name: str, start: int, size: int) -> IDASegment:
        # TODO: Support other segment class types.
        success = self._ida_segment.add_segm(0, start, start + size, name, "XTRN")
        if not success:
            raise ValueError(f"Unable to create segment at 0x{start:08x}")
        return self.get_segment(start)

    @property
    def segments(self) -> Iterable[IDASegment]:
        # Taken from idautils.Segments()
        for n in range(self._ida_segment.get_segm_qty()):
            seg = self._ida_segment.getnseg(n)
            if seg:
                yield IDASegment(self, seg)

    def get_string_bytes(self, addr: int, length: int = None, bit_width: int = None) -> bytes:
        if bit_width is None:
            str_type = self._idc.get_str_type(addr)
        elif bit_width == 8:
            str_type = self._ida_nalt.STRTYPE_C
        elif bit_width == 16:
            str_type = self._ida_nalt.STRTYPE_C_16
        elif bit_width == 32:
            str_type = self._ida_nalt.STRTYPE_C_32
        else:
            raise ValueError(f"Invalid bit width: {bit_width}")

        if length is None:
            length = self._ida_bytes.get_max_strlit_length(
                addr, str_type,
                self._ida_bytes.ALOPT_IGNCLT | self._ida_bytes.ALOPT_IGNPRINT | self._ida_bytes.ALOPT_MAX4K
            )

        return self._ida_bytes.get_strlit_contents(addr, length, str_type)

    def strings(self, min_length=3) -> Iterable[IDAString]:
        sc = self._idautils.Strings()
        sc.setup(
            strtypes=[
                self._ida_nalt.STRTYPE_C,
                self._ida_nalt.STRTYPE_C_16,
                self._ida_nalt.STRTYPE_C_32,
                self._ida_nalt.STRTYPE_PASCAL,
                self._ida_nalt.STRTYPE_PASCAL_16,
                self._ida_nalt.STRTYPE_LEN2,
                self._ida_nalt.STRTYPE_LEN2_16,
                self._ida_nalt.STRTYPE_LEN4,
                self._ida_nalt.STRTYPE_LEN4_16,
            ],
            minlen=min_length,
        )
        for string in sc:
            yield IDAString(self, string)

    def get_data_type(self, name: str) -> IDADataType:
        is_ptr = name.endswith("*")
        name = name.strip(" *")

        # Name has to be uppercase for get_named_type() to work.
        name = name.upper()

        # Create new tinfo object of type.
        tif = self._ida_typeinf.tinfo_t()
        success = tif.get_named_type(self._ida_typeinf.get_idati(), name)
        if not success:
            raise NotExistError(f"Invalid data type: {name}")

        # If a pointer, create another tinfo object that is the pointer of the first.
        if is_ptr:
            tif2 = self._ida_typeinf.tinfo_t()
            tif2.create_ptr(tif)
            tif = tif2

        return IDADataType(self, tif)

    @property
    def max_address(self) -> int:
        return self._ida_ida.inf_get_max_ea()

    @property
    def min_address(self) -> int:
        return self._ida_ida.inf_get_min_ea()

    @property
    def base_address(self) -> int:
        return self._ida_nalt.get_imagebase()

    def open_memory(self, start: int, end: int) -> IDAMemory:
        return IDAMemory(self, start, end)

    def references_from(self, addr: int) -> Iterable[IDAReference]:
        # TODO: Cache chunks
        for xref in self._idautils.XrefsFrom(addr):
            reference = IDAReference(self, xref)
            # Ignore "ordinary flow" references, since that's just a reference to the next
            # instruction, which Ghidra doesn't do.
            if reference.type == ReferenceType.ordinary_flow:
                continue
            yield reference

    def references_to(self, addr: int) -> Iterable[IDAReference]:
        # TODO: Cache chunks
        for xref in self._idautils.XrefsTo(addr):
            yield IDAReference(self, xref)

    def create_reference(self, from_address: int, to_address: int, ref_type: ReferenceType) -> IDAReference:
        ref_type = IDAReference._type_map_inv[ref_type]
        code = False
        if ref_type in range(1, 6):  # data reference
            success = self._ida_xref.add_dref(from_address, to_address, ref_type)
        elif ref_type in range(16, 22):  # code reference
            code = True
            success = self._ida_xref.add_cref(from_address, to_address, ref_type)
        else:
            raise ValueError(f"Unsupported reference type: {ref_type}")
        if not success:
            raise ValueError(f"Unable to create reference: 0x{from_address:08x} -> 0x{to_address:08x}")

        # IDA's xrefblt_t is actually an iterator which holds the attributes of the "current" reference
        # being iterated.
        # Copying the work done in xrefblt_t.refs_from()/refs_to()
        xref = self._ida_xref.xrefblk_t()
        xref.frm = from_address
        xref.to = to_address
        xref.type = ref_type
        xref.iscode = code
        xref.user = ref_type == 20
        reference = IDAReference(self, xref)

        # If reference is a new code call, tell analyzer to apply the callee's type to calling point.
        if reference.type == ReferenceType.code_call:
            self._ida_auto.auto_apply_type(reference.from_address, reference.to_address)

        return reference

    def get_variable(self, addr: int) -> IDAGlobalVariable:
        start_address = self._ida_bytes.get_item_head(addr)
        # Don't count code as "variables". Otherwise we get all the
        # loop labels as variables.
        flags = self._ida_bytes.get_flags(addr)
        is_code = self._ida_bytes.is_code(flags)
        # Only count as variable if item has a name.
        if not is_code and self._ida_name.get_name(start_address):
            return IDAGlobalVariable(self, start_address)
        else:
            raise NotExistError(f"Variable doesn't exist at {hex(addr)}")

    @property
    def imports(self) -> Iterable[IDAImport]:
        for address, thunk_address, name, namespace in self._ida_helpers.iter_imports():
            yield IDAImport(self, address, thunk_address, name, namespace)

    def get_import(self, name: str) -> IDAImport:
        # Using ida_helpers instead of default implementation to improve performance.
        ret = self._ida_helpers.get_import(name)
        if ret is None:
            raise NotExistError(f"Import with name '{name}' doesn't exist.")
        return IDAImport(self, *ret)

    @property
    def exports(self) -> Iterable[IDAExport]:
        ida_entry = self._ida_entry
        for i in range(ida_entry.get_entry_qty()):
            ordinal = ida_entry.get_entry_ordinal(i)
            address = ida_entry.get_entry(ordinal)
            name = ida_entry.get_entry_name(ordinal)
            yield IDAExport(self, address, name)


class IDALocal(IDAFlatAPI, IDALocalDisassembler):
    ...


class IDARemote(IDAFlatAPI, IDARemoteDisassembler):
    ...




from __future__ import annotations
import os
from typing import TYPE_CHECKING

from jpype.types import *

from dragodis.interface import Memory

if TYPE_CHECKING:
    import ghidra
    from dragodis.ghidra.flat import GhidraFlatAPI


class GhidraMemory(Memory):

    def __init__(self, ghidra: GhidraFlatAPI, start: int, end: int):
        super().__init__(start, end)
        self._ghidra = ghidra
        self._memory = self._ghidra._program.getMemory()

    def read(self, size: int = None) -> bytes:
        remaining_bytes = self.end - (self.start + self._offset)
        if size is None:
            size = remaining_bytes
        size = min(size, remaining_bytes)
        if not size:
            return b""

        address = self._ghidra._to_addr(self.start + self._offset)
        buffer = JByte[size]
        num_bytes = self._memory.getBytes(address, buffer)
        self._offset += num_bytes
        return bytes(buffer)

    def reset(self, size: int = None) -> int:
        remaining_bytes = self.end - (self.start + self._offset)
        if size is None:
            size = remaining_bytes
        size = min(size, remaining_bytes)
        if not size:
            return 0

        # Obtain original bytes directly from underlying file.
        address = self._ghidra._to_addr(self.start + self._offset)
        memory_block = self._memory.getBlock(address)
        memory_info = memory_block.getSourceInfos()[0]
        file_bytes = self._memory.getAllFileBytes()[0]
        file_offset = (
            address.getOffset()
            - memory_info.getMinAddress().getOffset()
            + memory_info.getFileBytesOffset()
        )
        buffer = JByte[size]
        file_bytes.getOriginalBytes(file_offset, buffer)

        # Write bytes.
        return self.write(bytes(buffer))

    def write(self, data: bytes) -> int:
        from ghidra.program.model.mem import MemoryAccessException
        # Trim given data to ensure we only write within the window.
        data = data[:self.end - (self.start + self._offset)]
        if not data:
            return 0
        address = self._ghidra._to_addr(self.start + self._offset)
        try:
            self._memory.setBytes(address, data)
        except MemoryAccessException as e:
            raise IOError(e)
        self._offset += len(data)
        return len(data)




from __future__ import annotations
from typing import List, TYPE_CHECKING, Optional

from dragodis.exceptions import NotExistError
from dragodis.ida.operand import IDAOperand, IDAx86Operand, IDAARMOperand
from dragodis.interface.types import FlowType
from dragodis.interface.instruction import (
    Instruction, x86Instruction, ARMInstruction, ARMConditionCode
)

if TYPE_CHECKING:
    from dragodis.ida.flat import IDAFlatAPI


# TODO: Perhaps have a local helper utility that pulls and caches all the instruction
#    objects for the function, ready to be accessed?
class IDAInstruction(Instruction):
    _Operand = IDAOperand

    def __init__(self, ida: IDAFlatAPI, addr: int):
        super().__init__(ida)
        self._ida = ida
        self._addr = addr
        insn_t = self._ida._ida_helpers.get_instruction(addr)
        if not insn_t:
            raise NotExistError(f"Instruction does not exist at: {hex(addr)}")
        self._insn_t = insn_t

    @property
    def address(self):
        return self._addr

    @property
    def flow_type(self) -> FlowType:
        if self._ida._ida_idp.is_call_insn(self._insn_t):
            return FlowType.call
        elif self._ida._ida_idp.is_ret_insn(self._insn_t):
            return FlowType.terminal

        refs = self._ida._idautils.CodeRefsFrom(self.address, 0)
        has_refs = bool(next(refs, False))
        if not has_refs:
            return FlowType.fall_through

        # Checking for conditional/unconditional jump is a bit more involved with IDA.
        # We are going to need to check if the code reference is a "flow" or not.
        next_head = self._ida._idc.next_head(self.address)
        if self._ida._idc.is_flow(self._ida._idc.get_full_flags(next_head)):
            return FlowType.conditional_jump
        else:
            return FlowType.unconditional_jump

    @property
    def mnemonic(self) -> str:
        return (self._ida._ida_ua.ua_mnem(self.address) or "").lower()

    @property
    def root_mnemonic(self) -> str:
        return self._insn_t.get_canon_mnem().lower()

    @property
    def operands(self) -> List[IDAOperand]:
        return [
            self._Operand(self, self._ida, self.address, index, op)
            for index, op in self._ida._ida_helpers.get_operands(self.address)
        ]

    @property
    def text(self) -> str:
        return self._ida._idc.GetDisasm(self.address)

    @property
    def stack_depth(self) -> int:
        return self._ida._idc.get_spd(self.address)

    @property
    def stack_delta(self) -> int:
        # NOTE: IDA gives the delta in relation to the previous instruction,
        #   but we want the delta that this instructions applies.
        delta = self._ida._idc.get_sp_delta(self._ida._idc.next_head(self.address))
        if delta is None:
            delta = 0
        return delta


class IDAx86Instruction(IDAInstruction, x86Instruction):
    _Operand = IDAx86Operand

    @property
    def rep(self) -> Optional[str]:
        if self.data[0] in (0xF2, 0xF3):
            text = self.text.lower()
            # Existence of byte prefix doesn't necessarily mean we have a rep prefix.
            # Double confirm by looking at the text itself.
            if text.startswith("rep"):
                rep, _, _ = text.partition(" ")
                return rep
        return None


class IDAARMInstruction(IDAInstruction, ARMInstruction):
    _Operand = IDAARMOperand

    @property
    def update_flags(self) -> bool:
        return bool(self._insn_t.auxpref & self._ida._ida_arm.aux_cond)

    @property
    def condition_code(self) -> ARMConditionCode:
        condition = self._ida._ida_arm.get_cond(self._insn_t)
        return ARMConditionCode(condition)

    @property
    def writeback(self) -> bool:
        return bool(
            self._insn_t.auxpref & (
                self._ida._ida_arm.aux_postidx
                | self._ida._ida_arm.aux_wback
                | self._ida._ida_arm.aux_wbackldm
            )
            or self.mnemonic in ("push", "pop")
        )

    @property
    def pre_indexed(self) -> bool:
        return self.writeback and not self._insn_t.auxpref & self._ida._ida_arm.aux_postidx

    @property
    def post_indexed(self) -> bool:
        return self.writeback and bool(self._insn_t.auxpref & self._ida._ida_arm.aux_postidx)


IDAInstruction._x86Instruction = IDAx86Instruction
IDAInstruction._ARMInstruction = IDAARMInstruction



from __future__ import annotations

from typing import TYPE_CHECKING

from dragodis.interface import Reference, ReferenceType

if TYPE_CHECKING:
    import ghidra
    from dragodis.ghidra.flat import GhidraFlatAPI


class GhidraReference(Reference):

    def __init__(self, ghidra: GhidraFlatAPI, reference: "ghidra.program.model.symbol.Reference"):
        self._ghidra = ghidra
        self._reference = reference

    @property
    def _ref_type(self) -> "ghidra.program.model.symbol.RefType":
        return self._reference.getReferenceType()

    @property
    def from_address(self) -> int:
        return self._reference.getFromAddress().getOffset()

    @property
    def is_code(self) -> bool:
        # TODO
        return not self.is_data

    @property
    def is_data(self) -> bool:
        return self._ref_type.isData()

    @property
    def to_address(self) -> int:
        return self._reference.getToAddress().getOffset()

    # TODO: Look into this further.
    #   - Need to better standardize reference types to match with both IDA and Ghidra.
    @property
    def type(self) -> ReferenceType:
        ref_type = self._ref_type
        if ref_type.isData():
            if ref_type.isRead():
                return ReferenceType.data_read
            elif ref_type.isWrite():
                return ReferenceType.data_write
            else:
                return ReferenceType.data  # TODO: Sometimes it is not read or write...
        elif ref_type.isCall():
            return ReferenceType.code_call
        elif ref_type.isJump():
            return ReferenceType.code_jump
        elif ref_type.isFallthrough():
            return ReferenceType.ordinary_flow
        else:
            return ReferenceType.unknown




# Processor types
# (pulled from https://hex-rays.com/products/ida/support/idadoc/618.shtml)
PROCESSOR_8086 = "8086"             # Intel 8086                   ┐
PROCESSOR_80286R = "80286r"         # Intel 80286 real mode        │
PROCESSOR_80286P = "80286p"         # Intel 80286 protected mode   │
PROCESSOR_80386R = "80386r"         # Intel 80386 real mode        │
PROCESSOR_80386P = "80386p"         # Intel 80386 protected mode   │
PROCESSOR_80486R = "80486r"         # Intel 80486 real mode        │
PROCESSOR_80486P = "80486p"         # Intel 80486 protected mode   │
PROCESSOR_80586R = "80586r"         # Intel Pentium & MMX real mode│ (IBM PC family)
PROCESSOR_80586P = "80586p"         # Intel Pentium & MMX prot mode│
PROCESSOR_80686P = "80686p"         # Intel Pentium Pro & MMX      │
PROCESSOR_K62 = "k62"               # AMD K6-2 with 3DNow!         │
PROCESSOR_P2 = "p2"                 # Intel Pentium II             │
PROCESSOR_P3 = "p3"                 # Intel Pentium III            │
PROCESSOR_ATHLON = "athlon"         # AMD K7                       │
PROCESSOR_P4 = "p4"                 # Intel Pentium 4              │
PROCESSOR_METAPC = "metapc"         # Disassemble all IBMPC opcodes┘
PROCESSOR_8085 = "8085"             # Intel 8085                   ┐
PROCESSOR_Z80 = "z80"               # Zilog 80                     │
PROCESSOR_Z180 = "z180"             # Zilog 180                    │ (Zilog 80 family)
PROCESSOR_Z380 = "z380"             # Zilog 380                    │
PROCESSOR_64180 = "64180"           # Hitachi HD64180              │
PROCESSOR_GB = "gb"                 # Gameboy                      ┘
PROCESSOR_Z8 = "z8"                 # Zilog 8                        (Zilog 8 family)
PROCESSOR_860XR = "860xr"           # Intel 860 XR                 ┐ (Intel 860 family)
PROCESSOR_860XP = "860xp"           # Intel 860 XP                 ┘
PROCESSOR_8051 = "8051"             # Intel 8051                     (Intel 51 family)
PROCESSOR_80196 = "80196"           # Intel 80196                  ┐ (Intel 80196 family)
PROCESSOR_80196NP = "80196NP"       # Intel 80196NP, NU            ┘
PROCESSOR_M6502 = "m6502"           # MOS 6502                     ┐ (MOS Technology 65xx family)
PROCESSOR_M65C02 = "m65c02"         # MOS 65c02                    ┘
PROCESSOR_PDP11 = "pdp11"           # DEC PDP/11                     (PDP family)
PROCESSOR_68000 = "68000"           # Motorola MC68000             ┐
PROCESSOR_68010 = "68010"           # Motorola MC68010             │
PROCESSOR_68020 = "68020"           # Motorola MC68020             │
PROCESSOR_68030 = "68030"           # Motorola MC68030             │
PROCESSOR_68040 = "68040"           # Motorola MC68040             │ (Motorola 680x0 family)
PROCESSOR_68330 = "68330"           # Motorola CPU32 (68330)       │
PROCESSOR_68882 = "68882"           # Motorola MC68020 with MC68882│
PROCESSOR_68851 = "68851"           # Motorola MC68020 with MC68851│
PROCESSOR_68020EX = "68020EX"       # Motorola MC68020 with both   │
PROCESSOR_COLFIRE = "colfire"       # Motorola ColdFire            │
PROCESSOR_68K = "68K"               # Motorola MC680x0 all opcodes ┘
PROCESSOR_6800 = "6800"             # Motorola MC68HC00            ┐
PROCESSOR_6801 = "6801"             # Motorola MC68HC01            │
PROCESSOR_6803 = "6803"             # Motorola MC68HC03            │
PROCESSOR_6301 = "6301"             # Hitachi HD 6301              │
PROCESSOR_6303 = "6303"             # Hitachi HD 6303              │
PROCESSOR_6805 = "6805"             # Motorola MC68HC05            │ (Motorola 8bit family)
PROCESSOR_6808 = "6808"             # Motorola MC68HC08            │
PROCESSOR_6809 = "6809"             # Motorola MC68HC09            │
PROCESSOR_6811 = "6811"             # Motorola MC68HC11            ┘
PROCESSOR_6812 = "6812"             # Motorola MC68HC12
PROCESSOR_HCS12 = "hcs12"           # Motorola MC68HCS12
PROCESSOR_6816 = "6816"             # Motorola MC68HC16
PROCESSOR_JAVA = "java"             # java                           (Java family)
PROCESSOR_PPC = "ppc"               # PowerPC big endian           ┐
PROCESSOR_PPCL = "ppcl"             # PowerPC little endian        ┘ (PowerPC family)
PROCESSOR_ARM = "arm"               # ARM little endian            ┐ (ARM family)
PROCESSOR_ARMB = "armb"             # ARM big endian               ┘
PROCESSOR_TMS320C2 = "tms320c2"     # TMS320C2x series             ┐ (TMS 16bit addressing family)
PROCESSOR_TMS320C5 = "tms320c5"     # TMS320C5x series             ┘
PROCESSOR_TMS320C6 = "tms320c6"     # TMS320C6x series             ┐ (TMS VLIW family)
PROCESSOR_TMS320C3 = "tms320c3"     # TMS320C3x series             ┘
PROCESSOR_TMS32054 = "tms32054"     # TMS320C54xx series
PROCESSOR_TMS32055 = "tms32055"     # TMS320C55xx series
PROCESSOR_SH3 = "sh3"               # Renesas SH-3 (little endian) ┐
PROCESSOR_SH3B = "sh3b"             # Renesas SH-3 (big endian)    │ (Renesas SuperH series)
PROCESSOR_SH4 = "sh4"               # Renesas SH-4 (little endian) │
PROCESSOR_SH4B = "sh4b"             # Renesas SH-4 (big endian)    │
PROCESSOR_SH2A = "sh2a"             # Renesas SH-2A (big endian)   ┘
PROCESSOR_AVR = "avr"               # ATMEL AVR                      (ATMEL family)
PROCESSOR_MIPSL = "mipsl"           # MIPS little endian           ┐ (MIPS family:R2000, R3000, R4000,
PROCESSOR_MIPSB = "mipsb"           # MIPS big endian              │ R4200, R4300,  R4400, R4600,
PROCESSOR_MIPSRL = "mipsrl"         # MIPS & RSP little            │ R8000, R10000)
PROCESSOR_MIPSR = "mipsr"           # MIPS & RSP big               │
PROCESSOR_R5900L = "r5900l"         # MIPS R5900 little            ┘
PROCESSOR_R5900R = "r5900r"         # MIPS R5900 big               ┘
PROCESSOR_H8300 = "h8300"           # H8/300x in normal mode       ┐
PROCESSOR_H8300A = "h8300a"         # H8/300x in advanced mode     │ (Hitachi H8 family)
PROCESSOR_H8S300 = "h8s300"         # H8S in normal mode           │
PROCESSOR_H8S300A = "h8s300a"       # H8S in advanced mode         ┘
PROCESSOR_H8500 = "h8500"           # H8/500                         (Hitachi H8/500 family)
PROCESSOR_PIC12CXX = "pic12cxx"     # Microchip PIC 12-bit (12xxx) ┐
PROCESSOR_PIC16CXX = "pic16cxx"     # Microchip PIC 14-bit (16xxx) │ (PIC family)
PROCESSOR_PIC18CXX = "pic18cxx"     # Microchip PIC 16-bit (18xxx) ┘
PROCESSOR_SPARCB = "sparcb"         # SPARC big endian             ┐
PROCESSOR_SPARCL = "sparcl"         # SPARC little endian          ┘ (SPARC family)
PROCESSOR_ALPHAB = "alphab"         # DEC Alpha big endian         ┐
PROCESSOR_ALPHAL = "alphal"         # DEC Alpha little endian      ┘ (ALPHA family)
PROCESSOR_HPPA = "hppa"             # HP PA-RISC big endian          (HP PA-RISC family)
PROCESSOR_DSP56K = "dsp56k"         # Motorola DSP 5600x           ┐
PROCESSOR_DSP561XX = "dsp561xx"     # Motorola DSP 561xx           │ (DSP 56K family)
PROCESSOR_DSP563XX = "dsp563xx"     # Motorola DSP 563xx           │
PROCESSOR_DSP566XX = "dsp566xx"     # Motorola DSP 566xx           ┘
PROCESSOR_C166 = "c166"             # Siemens C166                 ┐
PROCESSOR_C166V1 = "c166v1"         # Siemens C166 v1 family       │
PROCESSOR_C166V2 = "c166v2"         # Siemens C166 v2 family       │ (C166 family)
PROCESSOR_ST10 = "st10"             # SGS-Thomson ST10             │
PROCESSOR_SUPER10 = "super10"       # Super10                      ┘
PROCESSOR_ST20 = "st20"             # SGS-Thomson ST20/C1            (ST20 family)
PROCESSOR_ST20C4 = "st20c4"         # SGS-Thomson ST20/C2-C4
PROCESSOR_ST7 = "st7"               # SGS-Thomson ST7                (ST7 family)
PROCESSOR_IA64L = "ia64l"           # Intel Itanium little endian  ┐
PROCESSOR_IA64B = "ia64b"           # Intel Itanium big endian     ┘ (IA64 family)
PROCESSOR_CLI = "cli"               # Microsoft.Net platform
PROCESSOR_NET = "net"               # Microsoft.Net platform (alias)
PROCESSOR_I960L = "i960l"           # Intel 960 little endian      ┐
PROCESSOR_I960B = "i960b"           # Intel 960 big endian         ┘ (i960 family)
PROCESSOR_F2MC16L = "f2mc16l"       # Fujitsu F2MC-16L             ┐
PROCESSOR_F2MC16LX = "f2mc16lx"     # Fujitsu F2MC-16LX            ┘ (Fujitsu F2MC family)
PROCESSOR_78K0 = "78k0"             # NEC 78k/0
PROCESSOR_78K0S = "78k0s"           # NEC 78k/0s
PROCESSOR_M740 = "m740"             # Mitsubishi 8-bit
PROCESSOR_M7700 = "m7700"           # Mitsubishi 16-bit            ┐
PROCESSOR_M7750 = "m7750"           # Mitsubishi 16-bit            ┘ (Mitsubishi 16-bit family)
PROCESSOR_M32R = "m32r"             # Mitsubishi 32-bit            ┐
PROCESSOR_M32RX = "m32rx"           # Mitsubishi 32-bit extended   ┘ (Mitsubishi 32-bit family)
PROCESSOR_ST9 = "st9"               # STMicroelectronics ST9+
PROCESSOR_FR = "fr"                 # Fujitsu FR family
PROCESSOR_M7900 = "m7900"           # Mitsubishi M7900
PROCESSOR_KR1878 = "kr1878"         # Angstrem KR1878
PROCESSOR_AD218X = "ad218x"         # Analog Devices ADSP
PROCESSOR_OAKDSP = "oakdsp"         # Atmel OAK DSP
PROCESSOR_TRICORE = "tricore"       # Infineon Tricore
PROCESSOR_EBC = "ebc"               # EFI Bytecode
PROCESSOR_MSP430 = "msp430"         # Texas Instruments MSP430



"""
Interface for cross references.
"""

import abc

from dragodis.interface.types import ReferenceType


class Reference(metaclass=abc.ABCMeta):
    """
    References to represent the references to or from any address or function found in a disassembler.
    """

    def __eq__(self, other):
        return isinstance(other, Reference) and (
            self.from_address == other.from_address
            and self.to_address == other.to_address
            and self.type == other.type
        )

    def __hash__(self):
        return hash((self.from_address, self.to_address, self.type))

    def __str__(self) -> str:
        return f"{self.type.name}: 0x{self.from_address:08x} --> 0x{self.to_address:08x}"

    def __repr__(self) -> str:
        return f"<Reference {self.type!r}: 0x{self.from_address:08x} --> 0x{self.to_address:08x}>"

    @property
    @abc.abstractmethod
    def from_address(self) -> int:
        """
        The source address
        """

    @property
    @abc.abstractmethod
    def is_code(self) -> bool:
        """
        Is a reference to or from code.
        """

    @property
    @abc.abstractmethod
    def is_data(self) -> bool:
        ...

    @property
    def is_call(self) -> bool:
        """
        Is a code call reference.
        """
        return self.type == ReferenceType.code_call

    @property
    @abc.abstractmethod
    def to_address(self) -> int:
        """
        The destination reference
        """

    # TODO: have a is_call and is_jump type (see Sark)
    @property
    @abc.abstractmethod
    def type(self) -> ReferenceType:
        ...



"""
Interface for symbols
"""

from __future__ import annotations
import abc

from typing import Optional, Iterable, TYPE_CHECKING

from dragodis.interface.types import ReferenceType

if TYPE_CHECKING:
    from dragodis.interface import Reference


class Symbol(metaclass=abc.ABCMeta):
    """
    Symbols match a specific address to a string name.
    """

    def __str__(self) -> str:
        return f"{self.name}: 0x{self.address:08x}"

    def __repr__(self) -> str:
        return f"<Symbol {self}>"

    @property
    @abc.abstractmethod
    def address(self) -> int:
        """
        The address pointed to by the symbol
        """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        Name of the symbol
        """

    @property
    @abc.abstractmethod
    def references_to(self) -> Iterable[Reference]:
        """
        References to symbol
        """


class Import(Symbol):
    """
    Imports are a type of Symbol which have an external source or module.
    """

    def __str__(self) -> str:
        return f"{self.namespace}/{self.name}: 0x{self.address:08x}"

    def __repr__(self) -> str:
        return f"<Import {self}>"

    @property
    @abc.abstractmethod
    def namespace(self) -> Optional[str]:
        """
        The name of the external source of the import if available.
        Which is usually the name of the DLL.
        """

    @property
    @abc.abstractmethod
    def thunk_address(self) -> Optional[int]:
        """
        The address for the thunk function if it exists.
        """

    @property
    def calls_to(self) -> Iterable[int]:
        """
        Iterates addresses that call this import.
        """
        for ref in self.references_to:
            if ref.type == ReferenceType.code_call:
                yield ref.from_address


class Export(Symbol):
    """
    Exports are a type of Symbol which are declared entry points to the binary.
    """

    def __repr__(self) -> str:
        return f"<Export {self}>"




from __future__ import annotations
import abc
from typing import List, Iterable, Optional, TYPE_CHECKING, Tuple

from dragodis.exceptions import NotExistError
from dragodis.interface.function_signature import FunctionSignature
from dragodis.interface.stack import StackFrame
from dragodis.interface.types import ReferenceType

if TYPE_CHECKING:
    from dragodis.interface import Reference, Instruction, Flowchart, FlatAPI, Variable

from dragodis.interface.line import CommentType, Line


class Function(metaclass=abc.ABCMeta):
    """
    Function objects represent the actual functions that are found in a disassembler.
    """

    def __init__(self, api: FlatAPI):
        self._api = api

    def __hash__(self):
        return self.start

    def __eq__(self, other):
        return isinstance(other, Function) and self.start == other.start

    def __contains__(self, addr: int) -> bool:
        """
        Returns whether the given address is within the function.
        Defaults to checking if address is in-between start and end.
        """
        return self.start <= addr < self.end

    def __str__(self) -> str:
        return f"{self.name}()"

    def __repr__(self):
        return f"<Function 0x{self.start:08x}: {self}>"

    @property
    @abc.abstractmethod
    def start(self) -> int:
        """
        Returns the start address for the function.

        :return: The start address for the function
        :rtype: int
        """
        pass

    @property
    @abc.abstractmethod
    def end(self) -> int:
        """
        Returns the end address for the function.

        The end address is *not* part of the function.

        :return: The end address for the function
        :rtype: int
        """
        pass

    @property
    @abc.abstractmethod
    def flowchart(self) -> Flowchart:
        """
        The Flowchart that makes up the function.
        """

    @abc.abstractmethod
    def get_comment(self, comment_type=CommentType.anterior) -> Optional[str]:
        """
        Obtains comment for function.

        :param comment_type: Type of comment to get (defaults to end of line comment)
        :returns: String containing comment or None if there is no comment.
        """

    def lines(self, start: int = None, end: int = None, reverse=False) -> Iterable[Line]:
        """
        Iterates the line items in the function.

        NOTE: This is BFS using the flowchart.
        If you need something simpler you can use .lines() directly:

        .. code:: python

            lines = dis.lines(func.start, func.end)
        """
        for line in self.flowchart.lines(start, reverse=reverse):
            if end and line.address == end:
                break
            yield line

    def instructions(self, start: int = None, end: int = None, reverse=False) -> Iterable[Instruction]:
        """
        Iterates the instructions in the function.
        """
        for line in self.lines(start=start, end=end, reverse=reverse):
            insn = line.instruction
            if insn:
                yield insn

    @property
    def variables(self) -> Iterable[Variable]:
        """
        Iterates the variables in the function.
        """
        for insn in self.instructions():
            yield from insn.variables

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        Returns the name of the function.

        NOTE: The default name of a function is disassembler specific.
        """

    @name.setter
    @abc.abstractmethod
    def name(self, new_name: Optional[str]):
        """
        Sets the name of the function or resets the name if None/"" is provided.
        """

    @property
    def references_from(self) -> Iterable[Reference]:
        """
        Iterates external cross references from the function.
        This includes references for all instructions contained within the function.
        However, only references to external locations are counted.

        :yield: `Reference` objects.
        """
        for instruction in self.instructions():
            for ref in instruction.references_from:
                # Only count external references to avoid recursive loops.
                if ref.to_address not in self:
                    yield ref

    @property
    def references_to(self) -> Iterable[Reference]:
        """
        Iterates external cross references to the function.
        This includes references for all instructions contained within the function.
        However, only references from external locations are counted.

        :yield: `Reference` objects.
        """
        for instruction in self.instructions():
            for ref in instruction.references_to:
                # Only count external references to avoid recursive loops.
                if ref.from_address not in self:
                    yield ref

    @abc.abstractmethod
    def set_comment(self, comment: str, comment_type=CommentType.anterior):
        """
        Sets comment (of specific type) at line.

        :param str comment: comment string to set.
            (If value is an empty string, existing comments are cleared)
        :param comment_type: Type of comment to set (default to end of line comment)
        """

    @property
    def data(self) -> bytes:
        """
        Returns all the bytes contained in the function.
        WARNING: This doesn't yet properly support fragmented functions.
        """
        return self._api.get_bytes(self.start, self.end - self.start)

    @property
    def source_code(self) -> Optional[str]:
        """
        Provides decompiled source code for the function.

        WARNING: Implementing this property is optional since not all
            disassemblers may support decompilation.
            Users should handle the scenario when this property is None.

        WARNING: Decompiled source code is going to look very different based
            on what disassembler is used. Therefore, users should not use
            this in program logic and should really only be for display purposes.
        """
        return None

    @property
    @abc.abstractmethod
    def stack_frame(self) -> StackFrame:
        """
        Provides the stack frame for this function.
        """

    @property
    def signature(self) -> FunctionSignature:
        """
        The signature of the function.
        """
        return self._api.get_function_signature(self.start)

    @property
    def calls_to(self) -> Iterable[int]:
        """
        Iterates addresses that call this function.
        """
        for ref in self._api.references_to(self.start):
            if ref.type == ReferenceType.code_call:
                yield ref.from_address

    @property
    def calls_from(self) -> Iterable[Tuple[int, int]]:
        """
        Iterates call address and callee address for the calls within this function.
        """
        for ref in self.references_from:
            if ref.type == ReferenceType.code_call:
                yield ref.from_address, ref.to_address

    @property
    def callers(self) -> Iterable["Function"]:
        """
        Iterates Functions that call this function.
        NOTE: Recursive calls to itself are not counted.
        """
        cache = set()
        for addr in self.calls_to:
            try:
                func = self._api.get_function(addr)
            except NotExistError:
                continue
            if func.name not in cache:
                cache.add(func.name)
                yield func

    @property
    def callees(self) -> Iterable["Function"]:
        """
        Iterates Functions that this function calls.
        NOTE: Recursive calls to itself are not counted.
        """
        cache = set()
        for addr, func_addr in self.calls_from:
            try:
                func = self._api.get_function(func_addr)
            except NotExistError:
                continue
            if func.name not in cache:
                cache.add(func.name)
                yield func

    # TODO: Proper support for Function types.
    @property
    @abc.abstractmethod
    def is_library(self) -> bool:
        """
        Whether the function is a known library function.
        """





from __future__ import annotations
import abc
from functools import total_ordering
from typing import TYPE_CHECKING, Tuple

if TYPE_CHECKING:
    from dragodis.interface import Register


@total_ordering
class ArgumentLocation(metaclass=abc.ABCMeta):

    _PRIORITY: int

    def __lt__(self, other: ArgumentLocation) -> bool:
        return self._PRIORITY < other._PRIORITY

    @abc.abstractmethod
    def __eq__(self, other: ArgumentLocation) -> bool:
        ...


class StackLocation(ArgumentLocation):

    _PRIORITY = 3

    def __str__(self) -> str:
        return f"stack[0x{self.stack_offset:x}]"

    def __repr__(self) -> str:
        return f"<StackLocation 0x{self.stack_offset:x}>"

    @property
    @abc.abstractmethod
    def stack_offset(self) -> int:
        """
        The stack offset of the variable which holds the argument.
        NOTE: This offset is relative to the stack pointer BEFORE calling the function.
        It is the offset based on the current stack setup relative to the caller.
        ie. the first parameter would have an offset of 0 and second parameter would be +4
        on a 32bit x86 executable.
        """
        # TODO: Should we return a StackVariable object instead?

    def __lt__(self, other: ArgumentLocation) -> bool:
        if isinstance(other, StackLocation):
            return self.stack_offset < other.stack_offset
        return super().__lt__(other)

    def __eq__(self, other: ArgumentLocation) -> bool:
        if isinstance(other, StackLocation):
            return self.stack_offset == other.stack_offset
        return False


class RegisterLocation(ArgumentLocation):

    _PRIORITY = 0

    def __str__(self) -> str:
        return self.register.name

    def __repr__(self) -> str:
        return f"<RegisterLocation {self.register.name}>"

    @property
    @abc.abstractmethod
    def register(self) -> Register:
        """
        The register which holds the argument.
        """

    def __lt__(self, other: ArgumentLocation) -> bool:
        if isinstance(other, RegisterLocation):
            return self.register < other.register
        return super().__lt__(other)

    def __eq__(self, other: ArgumentLocation) -> bool:
        if isinstance(other, RegisterLocation):
            return self.register == other.register
        return False


class RegisterPairLocation(ArgumentLocation):

    _PRIORITY = 1

    def __str__(self) -> str:
        return f"({', '.join(reg.name for reg in self.registers)})"

    def __repr__(self) -> str:
        return f"<RegisterPairLocation {', '.join(repr(reg) for reg in self.registers)}>"

    @property
    @abc.abstractmethod
    def registers(self) -> Tuple[Register, Register]:
        """
        The pair of registers which holds the argument.
        """

    def __lt__(self, other: ArgumentLocation) -> bool:
        if isinstance(other, RegisterPairLocation):
            return self.registers < other.registers
        return super().__lt__(other)

    def __eq__(self, other: ArgumentLocation) -> bool:
        if isinstance(other, RegisterPairLocation):
            return self.registers == other.registers
        return False


class RelativeRegisterLocation(RegisterLocation):

    _PRIORITY = 2

    def __str__(self) -> str:
        return f"{self.register.name}[0x{self.offset:x}]"

    def __repr__(self) -> str:
        return f"<RelativeRegisterLocation {self.register!r}[0x{self.offset:x}]>"

    @property
    @abc.abstractmethod
    def offset(self) -> int:
        """
        The offset within the register which holds the argument.
        """

    def __lt__(self, other: ArgumentLocation) -> bool:
        if isinstance(other, RelativeRegisterLocation):
            return (self.register, self.offset) < (other.register, other.offset)
        return super().__lt__(other)

    def __eq__(self, other: ArgumentLocation) -> bool:
        if isinstance(other, RelativeRegisterLocation):
            return (self.register, self.offset) == (other.register, other.offset)
        return False


class StaticLocation(ArgumentLocation):

    _PRIORITY = 4

    def __str__(self) -> str:
        return f"0x{self.address:08x}"

    def __repr__(self) -> str:
        return f"<StaticLocation 0x{self.address:08x}>"

    @property
    @abc.abstractmethod
    def address(self) -> int:
        """
        The global address which holds the argument.
        """

    def __lt__(self, other: ArgumentLocation) -> bool:
        if isinstance(other, StaticLocation):
            return self.address < other.address
        return super().__lt__(other)

    def __eq__(self, other: ArgumentLocation) -> bool:
        if isinstance(other, StaticLocation):
            return self.address == other.address
        return False



"""
Interface for data types.
"""

import abc


class DataType(metaclass=abc.ABCMeta):

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"<DataType: {self.name}, size={self.size}>"

    def __eq__(self, other):
        return isinstance(other, DataType) and self.name == other.name and self.size == other.size

    @property
    def name(self) -> str:
        """
        The name of the data type.
        e.g: "dword", "char", etc.
        """

    @property
    def size(self) -> int:
        """
        The size of the data type.
        """



"""
Interface for cross references.
"""
from __future__ import annotations
from typing import TYPE_CHECKING

from dragodis.interface import Reference, ReferenceType

if TYPE_CHECKING:
    import ida_xref
    from dragodis.ida.flat import IDAFlatAPI


# noinspection PyPropertyAccess
class IDAReference(Reference):

    _type_map = {
        0: ReferenceType.unknown,          # ida_xref.fl_U
        1: ReferenceType.data_offset,           # ida_xref.dr_O
        2: ReferenceType.data_write,            # ida_xref.dr_W
        3: ReferenceType.data_read,             # ida_xref.dr_R
        4: ReferenceType.data_text,             # ida_xref.dr_T
        5: ReferenceType.data_informational,    # ida_xref.dr_I
        # 6: ReferenceType.??  # ida_xref.dr_S  # TODO
        16: ReferenceType.code_call,        # ida_xref.fl_CF
        17: ReferenceType.code_call,       # ida_xref.fl_CN
        18: ReferenceType.code_jump,        # ida_xref.fl_JF
        19: ReferenceType.code_jump,       # ida_xref.fl_JN
        20: ReferenceType.code_user,            # ida_xref.fl_USobsolete  # TODO: determine if we need this one?
        21: ReferenceType.ordinary_flow,        # ida_xref.fl_F
    }
    _type_map_inv = {value: key for key, value in _type_map.items()}

    def __init__(self, ida: IDAFlatAPI, xref: "ida_xref.xrefblk_t"):
        self._ida = ida
        self._xref = xref

    @property
    def from_address(self) -> int:
        return self._xref.frm

    @property
    def is_code(self) -> bool:
        return bool(self._xref.iscode)

    @property
    def is_data(self) -> bool:
        return self.type.name.startswith("data")  # TODO: confirm

    @property
    def to_address(self) -> int:
        return self._xref.to

    @property
    def type(self) -> ReferenceType:
        try:
            return self._type_map[self._xref.type]
        except KeyError:
            raise RuntimeError(f"Unexpected reference type: {self._xref.type}")



# Expose classes and functions as API.
from dragodis.api import open_program, IDA, Ghidra, Disassembler
from dragodis.exceptions import *
from dragodis.constants import *

# Import types from interface that may needed by users.
from dragodis.interface.types import *

__version__ = "0.7.1"




from .disassembler import BackendDisassembler
from .flat import FlatAPI
from .data_type import DataType
from .flowchart import Flowchart, BasicBlock
from .function import Function
from .function_signature import FunctionSignature, FunctionParameter
from .function_argument_location import (
    ArgumentLocation, StackLocation, RegisterLocation, RegisterPairLocation,
    RelativeRegisterLocation, StaticLocation,
)
from .instruction import Instruction
from .line import Line, LineType
from .memory import Memory
from .operand import Operand
from .operand_value import (
    OperandValue, Immediate, MemoryReference, Register, RegisterList, Phrase,
)
from .reference import Reference
from .segment import Segment
from .stack import StackFrame
from .symbol import Symbol
from .types import *
from .variable import *



"""
Utility functions used throughout the project.
"""
import contextlib
from typing import Optional

from elftools.common.exceptions import ELFError
from elftools.elf import elffile
import pefile
import pyhidra


class cached_property(property):
    """
    A cached_property which allows a setter.
    When the given property is set, this will automatically invalidate the cache.
    Good for read many / write rarely type properties in which the property value
    could only change if the setter was called.

    To invalidate the cache outside of setting the property use del.
        e.g. del obj.thing
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "__cached_" + self.fget.__name__

    def __get__(self, instance, type=None):
        """
        First attempts to pull from cached results before calling fget.
        """
        if instance is None:
            return self
        try:
            return instance.__dict__[self.name]
        except KeyError:
            res = instance.__dict__[self.name] = self.fget(instance)
            return res

    def __set__(self, instance, value):
        """
        Sets new value and invalidates cache.
        """
        self.fset(instance, value)
        self.__delete__(instance)

    def __delete__(self, instance):
        """
        Invalidates cache.
        """
        try:
            del instance.__dict__[self.name]
        except KeyError:
            pass


def is_64_bit(input_file) -> Optional[bool]:
    """
    Attempt to determine if the file is 64bit based on the pe header.
    Note that the pe.close() prevents an mmap'd file from
    being left open indefinitely as the PE object doesn't seem to get garbage collected.
    Forcing garbage collection also corrects that issue.

    :param input_file: The full path to the file in question

    :returns: True if 64 bit, False if not 64 bit or error in parsing the header.
    """
    # Get first bytes of file to check the file magic
    with open(input_file, "rb") as f:
        first_bytes = f.read(8)

    # PE file type
    if first_bytes[0:2] == b"\x4D\x5A":
        try:
            with contextlib.closing(pefile.PE(input_file, fast_load=True)) as pe:
                return pe.PE_TYPE == pefile.OPTIONAL_HEADER_MAGIC_PE_PLUS
        except pefile.PEFormatError:
            return False

    # elf file type
    elif first_bytes[1:4] == b"\x45\x4C\x46":
        try:
            with open(input_file, "rb") as f:
                elf = elffile.ELFFile(f)
                return elf.get_machine_arch() in ["AArch64", "x64"]
        except ELFError:
            return False

    # 32 bit MACH-O executable
    elif first_bytes[0:4] == b"\xCE\xFA\xED\xFE":
        return False

    # 64 bit MACH-O executable
    elif first_bytes[0:4] == b"\xCF\xFA\xED\xFE":
        return True

    return False


def in_ida() -> bool:
    """
    Detects if we are inside IDA.
    """
    try:
        import idc
        return True
    except (ModuleNotFoundError, ImportError):
        return False


def in_ghidra() -> bool:
    """
    Detects if we are inside a Ghidra interpreter with a program loaded.
    """
    interpreter = pyhidra.get_current_interpreter()
    return bool(interpreter and interpreter.currentProgram)



"""
This package contains miscellaneous helper functions that are to be run natively in IDA.
These are usually implemented here to improve performance or deal with weird
function callback mechanics.
(This is not a real module in the IDA SDK.)
"""
import logging
import re
from typing import Iterable, Tuple, List, Optional

import idc
import ida_ida
import ida_bytes
import ida_entry
import ida_funcs
import ida_hexrays
import ida_nalt
import ida_name
import ida_typeinf
import ida_ua
import idautils


logger = logging.getLogger(__name__)


def iter_imports(_target=None) -> Iterable[Tuple[int, Optional[int], str, str]]:
    """
    Iterates the imports, returning the address function name and namespace
    for the import.

    :param _target: Internally used to extract a specific import.

    :yields: (address, thunk_address or None, func_name, namespace)
    """
    for i in range(ida_nalt.get_import_module_qty()):
        namespace = ida_nalt.get_import_module_name(i)

        entries = []

        def callback(addr, name, ordinal):
            if name:
                # Name will include a "__imp_" prefix if the import is accessed through
                # a thunk function.
                # Pull the address of the thunk function as well.
                thunk_addr = None
                raw_name = ida_name.get_name(addr)
                if raw_name.startswith("__imp_"):
                    for xref in idautils.XrefsTo(addr):
                        func = ida_funcs.get_func(xref.frm)
                        if func and func.flags & ida_funcs.FUNC_THUNK:
                            # Pull thunk address and swap name for a better one without the "__imp_" prefix.
                            thunk_addr = func.start_ea
                            name = ida_funcs.get_func_name(thunk_addr)
                            break
                    else:
                        raise RuntimeError(f"Failed to find a thunk for {name} at 0x{addr:08X}")

                if _target and name != _target:
                    return True  # continue enumeration
                entries.append((addr, thunk_addr, name))
                if _target and name == _target:
                    return False  # target found, stop enumeration

            return True  # continue enumeration

        ida_nalt.enum_import_names(i, callback)

        for addr, thunk_addr, name in entries:
            yield addr, thunk_addr, name, namespace


def get_import(name: str) -> Optional[Tuple[int, int, str, str]]:
    """
    Gets import for given name.
    """
    for address, thunk_address, import_name, namespace in iter_imports(_target=name):
        return address, thunk_address, import_name, namespace


def iter_exports() -> Iterable[Tuple[int, str]]:
    """
    Iterate API exports.

    :yield: (ea, name)
    """
    for i in range(ida_entry.get_entry_qty()):
        ordinal = ida_entry.get_entry_ordinal(i)
        ea = ida_entry.get_entry(ordinal)
        name = ida_entry.get_entry_name(ordinal)
        yield ea, name


def get_byte_chunks(address: int, size: int) -> Iterable[Tuple[int, int]]:
    """
    Iterates the chunks of defined bytes found within the given address range.

    :param address: Starting address.
    :param size: Size of address range.
    :yields: Tuples of chunks of (address, size)
    """
    if size == 0:
        return
    elif size < 0:
        raise ValueError(f"Size must be positive. Got {size}")

    start = address
    end = start + size

    if not ida_bytes.is_loaded(address):
        address = ida_bytes.next_inited(address, end)

    chunk_start = address
    while address != idc.BADADDR:
        prev_address = address
        # Get the next address that has a byte with is_loaded() == True
        address = ida_bytes.next_inited(address, end)
        if address != prev_address + 1:
            # Found a hole.
            yield chunk_start, prev_address - chunk_start + 1
            chunk_start = address


def get_all_byte_chunks() -> Iterable[Tuple[int, int]]:
    """
    Iterates all chunks of defined bytes in the sample.
    """
    start = ida_ida.inf_get_min_ea()
    end = ida_ida.inf_get_max_ea()
    yield from get_byte_chunks(start, end - start)


def get_bytes(address: int, size: int, default: int = 0) -> bytes:
    """
    Obtains bytes from given address.
    Replaces non-loaded data with 0 byte.

    :param address: Address to pull bytes from.
    :param size: Number of bytes to pull.
    :param default: Default byte to provide if not loaded.
    :returns: obtained bytes
    """
    if default == 0:
        data = bytearray(size)
    else:
        data = bytearray([default] * size)

    start = address
    for address, size in get_byte_chunks(address, size):
        offset = address - start
        data[offset:offset + size] = ida_bytes.get_bytes(address, size)

    return bytes(data)


def revert_bytes(address: int, size: int):
    """
    Reverts patched bytes back to the underlying data.

    :param address: Address to start reverting bytes.
    :param size: Number of bytes to revert.
    """
    for offset in range(address, address + size):
        ida_bytes.revert_byte(offset)


def is_loaded(address: int, size: int) -> bool:
    """
    Checks if all bytes are loaded.

    :param address: Address of first byte.
    :param size: Number of bytes to check.
    """
    return all(ida_bytes.is_loaded(offset) for offset in range(address, address + size))


def get_instruction(address: int) -> ida_ua.insn_t:
    """
    Obtains insn_t object for instruction.
    """
    insn_t = ida_ua.insn_t()
    if ida_ua.decode_insn(insn_t, address):
        return insn_t


def get_operands(address: int) -> List[Tuple[int, ida_ua.op_t]]:
    """
    Obtains operands for instruction at given address.
    """
    insn = get_instruction(address)

    ret = []
    ops = insn.ops
    o_void = ida_ua.o_void
    for index, op in enumerate(ops):
        if op.type == o_void:
            break  # no more operands

        ret.append((index, op))

    return ret


def _get_tif_with_hex_rays(address: int) -> Optional[ida_typeinf.tinfo_t]:
    """
    Attempt to get the tinfo_t object for the function using the Hex-Rays decompiler plugin.

    :raises: RuntimeError on failure.
    :returns: tinfo_t object on success.
    """
    decompiled = decompiled_code(address)
    if not decompiled:
        return None
    tif = ida_typeinf.tinfo_t()
    decompiled.get_func_type(tif)
    return tif


def _get_tif_with_guess_type(address: int) -> Optional[ida_typeinf.tinfo_t]:
    """
    Attempt to get the tinfo_t object for the function using the "guess_type" function.

    :raises: RuntimeError on failure.
    :returns: tinfo_t object on success.
    """
    guessed_type = idc.guess_type(address)
    if guessed_type is None:
        return None

    func_name = idc.get_func_name(address)
    if func_name is None:
        return None

    # Documentation states the type must be ';' terminated, also the function name must be inserted
    guessed_type = re.sub(r"\(", f" {func_name}(", f"{guessed_type};")
    set_type_result = idc.SetType(address, guessed_type)
    if not set_type_result:
        logger.warning(
            f"Failed to SetType for function at 0x{address:X} with guessed type {guessed_type!r}"
        )

    tif = ida_typeinf.tinfo_t()
    if not ida_nalt.get_tinfo(tif, address):
        return None
    return tif


_seen_func_types = set()

def get_func_type_info(address: int, operand: Tuple[int, int] = None) -> Tuple[ida_typeinf.func_type_data_t, ida_typeinf.tinfo_t]:
    """
    Obtain a idaapi.func_type_data_t object for the function with the provided start address.

    :param address: start address of the function
    :param operand: Optional address and index pair for an operand containing the function address in its value.
        This can be provided when function is dynamically generated at runtime. (e.g. call eax)

    :return: ida_typeinf.func_type_data_t object, ida_typeinf.tinfo_t object

    :raise RuntimeError: if func_type_data_t object cannot be obtained
    """
    func_type = idc.get_type(address)

    # First see if it's a type we already set before.
    if func_type and address in _seen_func_types:
        tif = ida_typeinf.tinfo_t()
        ida_nalt.get_tinfo(tif, address)

    # Otherwise, try to use the Hexrays decompiler to determine function signature.
    # (It's better than IDA's guess_type)
    else:
        # First try to get type information from the decompiled code produced
        # by the Hex Rays plugin.
        tif = _get_tif_with_hex_rays(address)

        if not tif:
            # Otherwise, if IDA's disassembler set it already, go with that.
            if func_type:
                tif = ida_typeinf.tinfo_t()
                ida_nalt.get_tinfo(tif, address)

            # Finally, see if we can obtain it with guess_type()
            else:
                tif = _get_tif_with_guess_type(address)

    if tif:
        func_type_data = ida_typeinf.func_type_data_t()

        # In IDA 7.6, imported functions are now function pointers.
        # To handle this, check if we need to pull out a pointed object first
        if tif.is_funcptr():
            tif = tif.get_pointed_object()

        success = tif.get_func_details(func_type_data)
        if success:
            # record that we have processed this function before. (and that we can grab it from the offset)
            _seen_func_types.add(address)
            func_type_data, tif = apply_func_type_data(address, func_type_data)
            return func_type_data, tif

    # If we have still failed, we have one more trick under our sleeve.
    # Try to pull the type information from the operand of the call instruction.
    # This could be set if the function has been dynamically created.
    if operand:
        tif = ida_typeinf.tinfo_t()
        ida_nalt.get_op_tinfo(tif, operand.address, operand.index)
        func_type_data = ida_typeinf.func_type_data_t()
        success = tif.get_func_details(func_type_data)
        if success:
            func_type_data, tif = apply_func_type_data(address, func_type_data)
            return func_type_data, tif

    raise RuntimeError(f"Failed to obtain func_type_data_t object for offset 0x{address:X}")


def apply_func_type_data(address: int, func_type_data: ida_typeinf.func_type_data_t) -> Tuple[ida_typeinf.func_type_data_t, ida_typeinf.tinfo_t]:
    """
    Applies given func_type_data_t to function pointed by given address.

    :param address: Address of function.
    :param func_type_data: Type information to apply to function.

    :return: Tuple containing newly created func_type_data_t and tinfo_t objects. (Toss the old one)
    """
    # Create new tif and apply.
    tif = ida_typeinf.tinfo_t()
    success = tif.create_func(func_type_data)
    if not success:
        raise RuntimeError(f"Failed to create new function tinfo object.")
    success = ida_typeinf.apply_tinfo(address, tif, ida_typeinf.TINFO_DEFINITE)
    if not success:
        raise RuntimeError(f"Failed to apply function signatures changes.")

    # We also need to create a new func_type_data, because the old one gets
    # borked for some reason.
    func_type_data = ida_typeinf.func_type_data_t()
    success = tif.get_func_details(func_type_data)
    if not success:
        raise RuntimeError(f"Failed to generate new func_type_data object.")

    return func_type_data, tif


def decompiled_code(address: int, _visited=None) -> Optional[ida_hexrays.cfuncptr_t]:
    """
    Generates IDA object representing the decompiled code for the given address.

    :param address: Start address of the function.

    :returns: cfuncptr_t object or None on failure.
    """
    if _visited is None:
        _visited = set()

    attempted_before = address in _visited
    _visited.add(address)

    # This requires Hexrays decompiler, load it and make sure it's available before continuing.
    if not ida_hexrays.init_hexrays_plugin():
        idc.load_and_run_plugin("hexrays", 0) or idc.load_and_run_plugin("hexx64", 0)
    if not ida_hexrays.init_hexrays_plugin():
        logger.debug("Unable to load Hexrays decompiler.")
        return None
    fail_obj = ida_hexrays.hexrays_failure_t()
    try:
        code = ida_hexrays.decompile(address, fail_obj)
    except ida_hexrays.DecompilationFailure as e:
        logger.warning(f"Failed to decompile function: {e}")
        return None
    if code and not fail_obj.code:
        return code

    if not fail_obj:
        logger.warning(f"Unable to decompile function at {hex(address)}")
        return None

    # Cannot possibly recover from call analysis failure.
    if fail_obj.code == -12:
        logger.warning(f"Unable to decompile function at {hex(address)}: call analysis failed")
        return None

    # We may be able to still recover from this by first trying to decompile
    # the called function that caused the failure.
    # If we've attempted this before, ensure we don't try a third time
    # and cause an infinite loop.
    if not attempted_before:
        failed_address = fail_obj.errea
        if ida_ua.ua_mnem(failed_address) == "call":
            call_address = idc.get_operand_value(failed_address, 0)
            if decompiled_code(_visited=_visited) is not None:
                return decompiled_code(address, visited=_visited)

    # TODO: Observed this message pops up with fail_obj.code == 0... unsure if that is actually an error.
    logger.debug(f"Unable to decompile function at {hex(address)}: {fail_obj.code}")
    return None


def get_function_by_name(name: str, ignore_underscore: bool = False) -> Optional[ida_funcs.func_t]:
    """
    Gets function by name (if it exists.)
    """
    for ea in idautils.Functions():
        func_name = ida_funcs.get_func_name(ea)
        if ignore_underscore:
            func_name = func_name.strip("_")
        if func_name == name:
            return ida_funcs.get_func(ea)




from __future__ import annotations
from typing import TYPE_CHECKING, Union, Iterable

from dragodis.ghidra.variable import GhidraStackVariable
from dragodis.interface.stack import StackFrame

if TYPE_CHECKING:
    from dragodis.ghidra.flat import GhidraFlatAPI
    import ghidra


class GhidraStackFrame(StackFrame):

    def __init__(self, ghidra: GhidraFlatAPI, frame: "ghidra.program.model.listing.StackFrame"):
        self._ghidra = ghidra
        self._frame = frame

    def __eq__(self, other):
        return self is other or self._frame == other._frame

    def __getitem__(self, name_or_offset: Union[str, int]) -> GhidraStackVariable:
        """
        Obtains stack variable based on name or offset.
        """
        if isinstance(name_or_offset, int):
            offset = name_or_offset
            # Casting int to ensure we don't pass in a JLong
            var = self._frame.getVariableContaining(int(offset))
            if var is not None:
                return GhidraStackVariable(self._ghidra, var)
        elif isinstance(name_or_offset, str):
            name = name_or_offset
            for variable in self:
                if variable.name and variable.name == name:
                    return variable
        raise KeyError(f"Unable to find stack variable from: {name_or_offset}")

    # TODO: For now, stack frame will be read-only.
    # def __setitem__(self, key, value: GhidraVariable):
    #     from ghidra.program.model.symbol import SourceType
    #     try:
    #         v = self[key]
    #         v._variable = value
    #     except KeyError:
    #         if isinstance(key, int) and key == value.loc or isinstance(key, str) and key == value.name:
    #             v = self._frame.createVariable(value.loc, value.dt, SourceType.USER_DEFINED)
    #     if v:
    #         return v
    #     raise KeyError

    def __delitem__(self, name_or_offset: Union[str, int]):
        self._frame.clearVariable(self[name_or_offset].stack_offset)

    def __len__(self) -> int:
        return len(self._frame.getStackVariables())

    def __iter__(self) -> Iterable[GhidraStackVariable]:
        for var in self._frame.getStackVariables():
            yield GhidraStackVariable(self._ghidra, var)



"""
Custom script to run rpyc server.
This was done so we can properly control the shutdown of IDA after the server exits.
"""
import sys

import rpyc
from rpyc import SlaveService, OneShotServer
from rpyc.core.stream import NamedPipeStream


def main():
    """
    Initialize rpyc connection.
    """
    # Don't compress to improve speed.
    rpyc.core.channel.Channel.COMPRESSION_LEVEL = 0

    # Importing idc here so we can still import the server externally.
    import idc

    if not idc.ARGV[1:]:
        raise RuntimeError(f"No connection parameter provided.")

    if sys.platform == "win32":
        pipe_name = idc.ARGV[1]
        stream = NamedPipeStream.create_server(pipe_name)
        with rpyc.classic.connect_stream(stream) as srv:
            srv.serve_all()
    else:
        socket_path = idc.ARGV[1]
        server = OneShotServer(SlaveService, socket_path=socket_path, auto_register=False)
        server.start()

    idc.qexit(0)


if __name__ == "__main__":
    main()




import logging
import os
import pathlib
from typing import TYPE_CHECKING, Optional

import pyhidra

from dragodis import utils
from dragodis.exceptions import NotInstalledError
from dragodis.interface import BackendDisassembler
from dragodis.constants import BACKEND_GHIDRA

# Used for typing to help Pycharm give us autocompletion.
if TYPE_CHECKING:
    import ghidra

logger = logging.getLogger(__name__)


class GhidraDisassembler(BackendDisassembler):

    name = BACKEND_GHIDRA

    PROCESSOR_ARM = "ARM:LE:32:v8"
    PROCESSOR_ARM64 = "AARCH64:LE:64:v8A"
    PROCESSOR_X86 = "x86:LE:32:default"
    PROCESSOR_X64 = "x86:LE:64:default"

    def __init__(self, input_path, processor=None):
        super().__init__(input_path, processor=processor)
        # Build on start()
        self._flatapi = None     # type: Optional[ghidra.program.flatapi.FlatProgramAPI]
        self._program = None     # type: Optional[ghidra.program.database.ProgramDB]
        self._listing = None     # type: Optional[ghidra.program.model.listing.Listing]
        # Built on demand
        self.__decomp_api = None
        self.__basic_block_model = None
        self.__monitor = None

    # region Helper Functions

    def stop(self, *exc_info):
        if self.__decomp_api is not None:
            # don't create one just to destroy it
            self._decomp_api.dispose()

    @property
    def _decomp_api(self) -> "ghidra.app.decompiler.flatapi.FlatDecompilerAPI":
        if not self.__decomp_api:
            from ghidra.app.decompiler.flatapi import FlatDecompilerAPI
            self.__decomp_api = FlatDecompilerAPI(self._flatapi)
            self.__decomp_api.initialize()
        return self.__decomp_api

    @property
    def _basic_block_model(self) -> "ghidra.program.model.block.BasicBlockModel":
        if not self.__basic_block_model:
            from ghidra.program.model.block import BasicBlockModel
            self.__basic_block_model = BasicBlockModel(self._program)
        return self.__basic_block_model

    @property
    def _monitor(self) -> "ghidra.util.task.TaskMonitor":
        if not self.__monitor:
            from ghidra.util.task import TaskMonitor
            self.__monitor = TaskMonitor.DUMMY
        return self.__monitor

    # endregion


class GhidraRemoteDisassembler(GhidraDisassembler):

    def __init__(self, input_path, ghidra_path=None, processor=None, **unused):
        """
        Initializes Ghidra disassembler.

        :param input_path: Path of binary to process.
        :param ghidra_path: Path to Ghidra directory.
            This may also be set using the environment variable GHIDRA_INSTALL_DIR
        """
        super().__init__(input_path, processor=processor)
        ghidra_path = ghidra_path or os.environ.get("GHIDRA_INSTALL_DIR", os.environ.get("GHIDRA_DIR"))
        if not ghidra_path:
            raise NotInstalledError(
                "Failed to get Ghidra install directory. "
                "Please provide it during instantiation or set the GHIDRA_INSTALL_DIR environment variable."
            )
        # Set environment variable for pyhidra.
        os.environ["GHIDRA_INSTALL_DIR"] = ghidra_path

        self._running = False
        self._bridge = None

    def start(self):
        if self._running:
            raise ValueError(f"Ghidra disassembler already running.")

        logger.debug(f"Starting pyhidra connection to {self.input_path}")
        self._bridge = pyhidra.open_program(self.input_path, language=self._processor)
        self._flatapi = self._bridge.__enter__()
        self._program = self._flatapi.getCurrentProgram()
        self._listing = self._program.getListing()

        self._running = True
        logger.debug("Ghidra Disassembler ready!")

    def stop(self, *exc_info):
        if not self._running:
            return
        logger.debug("Shutting down Ghidra connection...")
        super().stop(*exc_info)
        self._bridge.__exit__(*exc_info)
        self._bridge = None
        self._running = False
        logger.debug("Ghidra connection closed.")


class GhidraLocalDisassembler(GhidraDisassembler):

    def __init__(self, input_path=None, currentProgram=None, **unused):
        if not currentProgram:
            interpreter = pyhidra.get_current_interpreter()
            currentProgram = interpreter.currentProgram

        super().__init__(currentProgram.executablePath)
        self._program = currentProgram
        # Input path is not required when inside Ghidra, but if provided
        # let's use it to validate we are looking at the right file.
        if input_path and pathlib.Path(input_path).resolve() != self.input_path:
            raise ValueError(
                f"Expected input path isn't the same as the file loaded in Ghidra: {input_path} != {self.input_path}"
            )
        self.start()

    def start(self):
        self._listing = self._program.getListing()
        from ghidra.program.flatapi import FlatProgramAPI
        self._flatapi = FlatProgramAPI(self._program)




from __future__ import annotations
from typing import Tuple, TYPE_CHECKING

from dragodis.ghidra.operand_value import GhidraRegister
from dragodis.interface.function_argument_location import (
    ArgumentLocation, StackLocation, RegisterLocation,
    RegisterPairLocation, RelativeRegisterLocation, StaticLocation
)

if TYPE_CHECKING:
    import ghidra


class GhidraArgumentLocation(ArgumentLocation):

    def __init__(self, function: "ghidra.program.model.listing.Function", storage: "ghidra.program.model.listing.VariableStorage"):
        self._function = function
        self._storage = storage


class GhidraStackLocation(StackLocation, GhidraArgumentLocation):

    @property
    def stack_offset(self) -> int:
        # Ghidra returns a stack offset relative to the beginning of the callee.
        # ie. the return address would have been pushed in making the offset for
        # the first parameter at offset +4
        # Therefore, we will subtract the return address size if applicable.
        calling_conv = self._function.getCallingConvention()
        if calling_conv is None:
            fman = self._function.getProgram().getFunctionManager()
            calling_conv = fman.getDefaultCallingConvention()
        return self._storage.getStackOffset() - calling_conv.getStackshift()


class GhidraRegisterLocation(RegisterLocation, GhidraArgumentLocation):

    @property
    def register(self) -> GhidraRegister:
        return GhidraRegister(self._storage.getRegister())


class GhidraRegisterPairLocation(RegisterPairLocation, GhidraArgumentLocation):

    @property
    def registers(self) -> Tuple[GhidraRegister, GhidraRegister]:
        reg1, reg2 = self._storage.getRegisters()
        return GhidraRegister(reg1), GhidraRegister(reg2)


class GhidraRelativeRegisterLocation(RelativeRegisterLocation, GhidraArgumentLocation):

    @property
    def register(self) -> GhidraRegister:
        return GhidraRegister(self._storage.getRegister())

    @property
    def offset(self) -> int:
        return self._storage.getLastVarnode().getOffset()


class GhidraStaticLocation(StaticLocation, GhidraArgumentLocation):

    @property
    def address(self) -> int:
        return self._storage.getFirstVarnode().getAddress().getOffset()



"""
String Items
"""

import abc


class String(metaclass=abc.ABCMeta):
    """Found string item"""

    def __repr__(self):
        return f"<String 0x{self.address:08x}: '{str(self)}'>"

    def __str__(self):
        return self.value

    def __bytes__(self):
        return self.data

    @property
    @abc.abstractmethod
    def value(self) -> str:
        """
        The string found.
        """

    @property
    @abc.abstractmethod
    def data(self) -> bytes:
        """
        The raw bytes that make up the string.
        """

    @property
    @abc.abstractmethod
    def address(self) -> int:
        """
        The address of the string.
        """
